# -*- coding: utf-8 -*-
'''Test WGC surface intercpt point calculation.'''

import pytest

from webgeocalc import SurfaceInterceptPoint
from webgeocalc.errors import CalculationInvalidAttr, CalculationUndefinedAttr

@pytest.fixture
def kernels():
    '''Cassini kernel set.'''
    return 5

@pytest.fixture
def time():
    '''Input time.'''
    return '2012-10-19T08:24:00.000'

@pytest.fixture
def target():
    '''Input target.'''
    return 'ENCELADUS'

@pytest.fixture
def target_frame():
    '''Input target frame.'''
    return 'IAU_ENCELADUS'

@pytest.fixture
def observer():
    '''Input observer.'''
    return 'CASSINI'

@pytest.fixture
def direction_vector_type_1():
    '''Input direction vector type 1.'''
    return 'INSTRUMENT_BORESIGHT'

@pytest.fixture
def direction_vector_type_2():
    '''Input direction vector type 2.'''
    return 'REFERENCE_FRAME_AXIS'

@pytest.fixture
def direction_vector_type_3():
    '''Input direction vector type 3.'''
    return 'VECTOR_IN_INSTRUMENT_FOV'

@pytest.fixture
def direction_vector_type_4():
    '''Input direction vector type 4.'''
    return 'VECTOR_IN_REFERENCE_FRAME'

@pytest.fixture
def direction_instrument():
    '''Input direction instrument.'''
    return 'CASSINI_ISS_NAC'

@pytest.fixture
def direction_frame():
    '''Input direction frame.'''
    return 'CASSINI_ISS_NAC'

@pytest.fixture
def direction_frame_axis():
    '''Input direction axis.'''
    return 'Z'

@pytest.fixture
def corr():
    '''Input aberration correction.'''
    return 'NONE'

@pytest.fixture
def state():
    '''Input state.'''
    return 'LATITUDINAL'

@pytest.fixture
def params_1(kernels, time, target, target_frame, observer, direction_vector_type_1,
             direction_instrument, corr, state):
    '''Input parameters from WGC API example 1.'''
    return {
        'kernels': kernels,
        'times': time,
        'target': target,
        'target_frame': target_frame,
        'observer': observer,
        'direction_vector_type': direction_vector_type_1,
        'direction_instrument': direction_instrument,
        'aberration_correction': corr,
        'state_representation': state,
    }

@pytest.fixture
def payload_1(kernels, time, target, target_frame, observer, direction_vector_type_1,
              direction_instrument, corr, state):
    '''Payload from WGC API example 1.'''
    return {
        "kernels": [{
            "type": "KERNEL_SET",
            "id": kernels,
        }],
        "timeSystem": "UTC",
        "timeFormat": "CALENDAR",
        "times": [
            time,
        ],
        "calculationType": "SURFACE_INTERCEPT_POINT",
        "target": target,
        "shape1": "ELLIPSOID",
        "targetFrame": target_frame,
        "observer": observer,
        "directionVectorType": direction_vector_type_1,
        "directionInstrument": direction_instrument,
        "aberrationCorrection": corr,
        "stateRepresentation": state,
    }

@pytest.fixture
def params_2(kernels, time, target, target_frame, observer, direction_vector_type_2,
             direction_frame, direction_frame_axis, corr):
    '''Input parameters from WGC API example 2.'''
    return {
        'kernels': kernels,
        'times': time,
        'target': target,
        'target_frame': target_frame,
        'observer': observer,
        'direction_vector_type': direction_vector_type_2,
        'direction_frame': direction_frame,
        'direction_frame_axis': direction_frame_axis,
        'aberration_correction': corr,
    }

@pytest.fixture
def payload_2(kernels, time, target, target_frame, observer, direction_vector_type_2,
              direction_frame, direction_frame_axis, corr):
    '''Payload from WGC API example 2.'''
    return {
        "kernels": [{
            "type": "KERNEL_SET",
            "id": kernels,
        }],
        "timeSystem": "UTC",
        "timeFormat": "CALENDAR",
        "times": [
            time,
        ],
        "calculationType": "SURFACE_INTERCEPT_POINT",
        "target": target,
        "shape1": "ELLIPSOID",
        "targetFrame": target_frame,
        "observer": observer,
        "directionVectorType": direction_vector_type_2,
        "directionFrame": direction_frame,
        "directionFrameAxis": direction_frame_axis,
        "aberrationCorrection": corr,
        "stateRepresentation": "RECTANGULAR",
    }

@pytest.fixture
def params_3(kernels, time, target, target_frame, observer, direction_vector_type_3,
             direction_instrument, corr):
    '''Input parameters from WGC API example 3.'''
    return {
        'kernels': kernels,
        'times': time,
        'target': target,
        'target_frame': target_frame,
        'observer': observer,
        'direction_vector_type': direction_vector_type_3,
        "direction_instrument": direction_instrument,
        'direction_vector_x': 0,
        'direction_vector_y': 0,
        'direction_vector_z': 1,
        'aberration_correction': corr,
    }

@pytest.fixture
def payload_3(kernels, time, target, target_frame, observer, direction_vector_type_3,
              direction_instrument, corr):
    '''Payload from WGC API example 3.'''
    return {
        "kernels": [{
            "type": "KERNEL_SET",
            "id": kernels,
        }],
        "timeSystem": "UTC",
        "timeFormat": "CALENDAR",
        "times": [
            time,
        ],
        "calculationType": "SURFACE_INTERCEPT_POINT",
        "target": target,
        "shape1": "ELLIPSOID",
        "targetFrame": target_frame,
        "observer": observer,
        "directionVectorType": direction_vector_type_3,
        "directionInstrument": direction_instrument,
        "directionVectorX": 0,
        "directionVectorY": 0,
        "directionVectorZ": 1,
        "aberrationCorrection": corr,
        "stateRepresentation": "RECTANGULAR",
    }

@pytest.fixture
def params_4(kernels, time, target, target_frame, observer, direction_vector_type_4,
             direction_frame, corr):
    '''Input parameters from WGC API example 4.'''
    return {
        'kernels': kernels,
        'times': time,
        'target': target,
        'target_frame': target_frame,
        'observer': observer,
        'direction_vector_type': direction_vector_type_4,
        "direction_frame": direction_frame,
        'direction_vector_ra': 0,
        'direction_vector_dec': 90,
        'aberration_correction': corr,
    }

@pytest.fixture
def payload_4(kernels, time, target, target_frame, observer, direction_vector_type_4,
              direction_frame, corr):
    '''Payload from WGC API example 4.'''
    return {
        "kernels": [{
            "type": "KERNEL_SET",
            "id": kernels,
        }],
        "timeSystem": "UTC",
        "timeFormat": "CALENDAR",
        "times": [
            time,
        ],
        "calculationType": "SURFACE_INTERCEPT_POINT",
        "target": target,
        "shape1": "ELLIPSOID",
        "targetFrame": target_frame,
        "observer": observer,
        "directionVectorType": direction_vector_type_4,
        "directionFrame": direction_frame,
        "directionVectorRA": 0,
        "directionVectorDec": 90,
        "aberrationCorrection": corr,
        "stateRepresentation": "RECTANGULAR",
    }


def test_surface_direction_point_payload_boresight(params_1, payload_1):
    '''Test surface direction point payload with INSTRUMENT_BORESIGHT.'''
    assert SurfaceInterceptPoint(**params_1).payload == payload_1

def test_surface_direction_point_payload_reference_frame_axis(params_2, payload_2):
    '''Test surface direction point payload with REFERENCE_FRAME_AXIS.'''
    assert SurfaceInterceptPoint(**params_2).payload == payload_2

def test_surface_direction_point_payload_vector_in_instrument_fov(params_3, payload_3):
    '''Test surface direction point payload with VECTOR_IN_INSTRUMENT_FOV.'''
    assert SurfaceInterceptPoint(**params_3).payload == payload_3

def test_surface_direction_point_payload_vector_in_reference_frame(params_4, payload_4):
    '''Test surface direction point payload with VECTOR_IN_REFERENCE_FRAME.'''
    assert SurfaceInterceptPoint(**params_4).payload == payload_4

def test_surface_direction_point_attr_error(params_1):
    '''Test errors when surface direction point attributes is invalid.'''
    with pytest.raises(CalculationInvalidAttr):
        SurfaceInterceptPoint(shape_1='POINT', **params_1)

def test_surface_direction_point_payload_vector_in_instrument_fov_err(params_3):
    '''Test errors when VECTOR_IN_INSTRUMENT_FOV is invalid.'''
    del params_3['direction_vector_x']
    with pytest.raises(CalculationUndefinedAttr):
        # Missing 'direction_vector_x'
        SurfaceInterceptPoint(**params_3)



# -*- coding: utf-8 -*-
'''Configuration file for the Sphinx documentation builder.

This file does only contain a selection of the most common options. For a
full list see the documentation:
http://www.sphinx-doc.org/en/master/config
'''

import doctest

# -- Project information -----------------------------------------------------

project = 'WebGeoCalc API'
copyright = '2019, Benoit Seignovert'  # pylint: disable=W0622
author = 'Benoit Seignovert'

# -- General configuration ---------------------------------------------------

extensions = [
    'pbr.sphinxext',
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.doctest',
    'sphinx.ext.coverage',
    'sphinx.ext.viewcode',
]

source_suffix = '.rst'
master_doc = 'index'
language = None
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# -- Options for HTML output -------------------------------------------------

html_theme = 'sphinx_rtd_theme'
htmlhelp_basename = 'WebGeoCalcAPIdoc'

# -- Extension configuration -------------------------------------------------
autodoc_default_options = {
    'members': True,
    'show-inheritance': True,
}

autodoc_member_order = 'bysource'
pygments_style = None

doctest_default_flags = (
    0
    | doctest.DONT_ACCEPT_TRUE_FOR_1    # noqa: W503
    | doctest.ELLIPSIS                  # noqa: W503
    | doctest.IGNORE_EXCEPTION_DETAIL   # noqa: W503
    | doctest.NORMALIZE_WHITESPACE      # noqa: W503
)



# -*- coding: utf-8 -*-
'''Test WGC state vector calculation.'''

import pytest

from webgeocalc import StateVector

@pytest.fixture
def kernels():
    '''Cassini kernel set.'''
    return 5

@pytest.fixture
def target():
    '''Input target.'''
    return 'CASSINI'

@pytest.fixture
def observer():
    '''Input observer.'''
    return 'SATURN'

@pytest.fixture
def frame():
    '''Input frame.'''
    return 'IAU_SATURN'

@pytest.fixture
def time():
    '''Input time.'''
    return '2012-10-19T08:24:00.000'

@pytest.fixture
def corr():
    '''Input aberration correction.'''
    return 'NONE'

@pytest.fixture
def state():
    '''Input state.'''
    return 'PLANETOGRAPHIC'

@pytest.fixture
def params(kernels, target, observer, frame, time, corr, state):
    '''Input parameters from WGC API example.'''
    return {
        'kernels': kernels,
        'target': target,
        'observer': observer,
        'reference_frame': frame,
        'times': time,
        'aberration_correction': corr,
        'state_representation': state,
    }

@pytest.fixture
def payload(kernels, target, observer, frame, time, corr, state):
    '''Payload from WGC API example.'''
    return {
        "kernels": [
            {
                "type": "KERNEL_SET",
                "id": kernels
            }
        ],
        "timeSystem": "UTC",
        "timeFormat": "CALENDAR",
        "times": [
            time
        ],
        "calculationType": "STATE_VECTOR",
        "target": target,
        "observer": observer,
        "referenceFrame": frame,
        "aberrationCorrection": corr,
        "stateRepresentation": state
    }

def test_state_vector_payload(params, payload):
    '''Test state vector payload.'''
    assert StateVector(**params).payload == payload



# -*- coding: utf-8 -*-
'''Test WGC sub observer point calculation.'''

import pytest

from webgeocalc import SubObserverPoint

@pytest.fixture
def kernels():
    '''Cassini kernel set.'''
    return 5

@pytest.fixture
def time():
    '''Input time.'''
    return '2012-10-19T08:24:00.000'

@pytest.fixture
def target():
    '''Input target.'''
    return 'ENCELADUS'

@pytest.fixture
def target_frame():
    '''Input target frame.'''
    return 'IAU_ENCELADUS'

@pytest.fixture
def observer():
    '''Input observer.'''
    return 'CASSINI'

@pytest.fixture
def corr():
    '''Input aberration correction.'''
    return 'CN+S'

@pytest.fixture
def params(kernels, time, target, target_frame, observer, corr):
    '''Input parameters from WGC API example.'''
    return {
        'kernels': kernels,
        'times': time,
        'target': target,
        'target_frame': target_frame,
        'observer': observer,
        'aberration_correction': corr,
    }

@pytest.fixture
def payload(kernels, time, target, target_frame, observer, corr):
    '''Payload from WGC API example.'''
    return {
        "kernels": [{
            "type": "KERNEL_SET",
            "id": kernels,
        }],
        "timeSystem": "UTC",
        "timeFormat": "CALENDAR",
        "times": [
            time,
        ],
        "calculationType": "SUB_OBSERVER_POINT",
        "target": target,
        "targetFrame": target_frame,
        "observer": observer,
        "subPointType": "Near point: ellipsoid",
        "aberrationCorrection": corr,
        "stateRepresentation": "RECTANGULAR",
    }

def test_sub_observer_point_payload(params, payload):
    '''Test sub observer point payload.'''
    assert SubObserverPoint(**params).payload == payload



# -*- coding: utf-8 -*-
'''Command line interface.'''

import argparse
import re

from .api import API
from .calculation_types import (AngularSeparation, AngularSize,
                                Calculation, FrameTransformation,
                                IlluminationAngles, OsculatingElements,
                                StateVector, SubObserverPoint, SubSolarPoint,
                                SurfaceInterceptPoint, TimeConversion)
from .errors import KernelSetNotFound, TooManyKernelSets

def cli_kernel_sets(argv=None):
    '''Get list of kernel sets available on the server.

    GET: /kernel-sets
    '''
    parser = argparse.ArgumentParser(
        description='List and search kernel sets available in WebGeoCalc API.')
    parser.add_argument('--all', '-a', action='store_true',
                        help='List all kernel sets available')
    parser.add_argument('--kernel', '-k', metavar='NAME|ID', nargs='+',
                        help='Search a specific kernel by name or id')
    args, _ = parser.parse_known_args(argv)

    if args.all:
        kernels = API.kernel_sets()
        print('\n'.join([f" - {str(kernel)}: (id: {int(kernel)})" for kernel in kernels]))

    elif args.kernel:
        for kernel in args.kernel:
            try:
                kernel = int(kernel)
            except ValueError:
                pass

            try:
                kernel = API.kernel_set(kernel)
                print(f" - {str(kernel)}: (id: {int(kernel)})")

            except (KernelSetNotFound, TooManyKernelSets) as err:
                print(err)
    else:
        parser.print_help()


def cli_bodies(argv=None):
    '''Get list of bodies available in a kernel set.

    GET: /kernel-set/{kernelSetId}/bodies
    '''
    parser = argparse.ArgumentParser(
        description='List bodies available in WebGeoCalc API for a specific kernel set.')
    parser.add_argument('kernel', nargs='?', help='Kernel set name or id')
    parser.add_argument('--name', '-n', metavar='BODY', nargs='+',
                        help='Search a specific body by name or id')

    args, _ = parser.parse_known_args(argv)

    if args.kernel:
        try:
            kernel = int(args.kernel)
        except ValueError:
            kernel = args.kernel

        bodies = API.bodies(kernel)

        if args.name:
            for body in bodies:
                for name in args.name:
                    if name in body:
                        print(f" - {str(body)}: (id: {int(body)})")
        else:
            print('\n'.join([f" - {str(body)}: (id: {int(body)})" for body in bodies]))
    else:
        parser.print_help()

def cli_frames(argv=None):
    '''Get list of frames available in a kernel set.

    GET: /kernel-set/{kernelSetId}/frames
    '''
    parser = argparse.ArgumentParser(
        description='List frames available in WebGeoCalc API for a specific kernel set.')
    parser.add_argument('kernel', nargs='?', help='Kernel set name or id')
    parser.add_argument('--name', '-n', metavar='FRAME',
                        nargs='+', help='Search a specific frame by name or id')

    args, _ = parser.parse_known_args(argv)

    if args.kernel:
        try:
            kernel = int(args.kernel)
        except ValueError:
            kernel = args.kernel

        frames = API.frames(kernel)

        if args.name:
            for frame in frames:
                for name in args.name:
                    if name in frame:
                        print(f" - {str(frame)}: (id: {int(frame)})")
        else:
            print('\n'.join([f" - {str(frame)}: (id: {int(frame)})" for frame in frames]))
    else:
        parser.print_help()

def cli_instruments(argv=None):
    '''Get list of instruments available in a kernel set.

    GET: /kernel-set/{kernelSetId}/instruments
    '''
    parser = argparse.ArgumentParser(description='List instruments available'
                                     'in WebGeoCalc API for a specific kernel set.')
    parser.add_argument('kernel', nargs='?', help='Kernel set name or id')
    parser.add_argument('--name', '-n', metavar='INSTRUMENT', nargs='+',
                        help='Search a specific instrument by name or id')

    args, _ = parser.parse_known_args(argv)

    if args.kernel:
        try:
            kernel = int(args.kernel)
        except ValueError:
            kernel = args.kernel

        instruments = API.instruments(kernel)

        if args.name:
            for instrument in instruments:
                for name in args.name:
                    if name in instrument:
                        print(f" - {str(instrument)}: (id: {int(instrument)})")
        else:
            print('\n'.join(
                [f" - {str(instrument)}: (id: {int(instrument)})"
                 for instrument in instruments]))
    else:
        parser.print_help()

def _split(string, sep=','):
    # Replace and split string
    for char in ['[', ']', '=', '"', "'"]:
        string = string.replace(char, '')
    return string.split(sep)

def _underscore_case(string):
    # Convert string to underscore case (ie. snake case)
    string = string.replace('-', '_')
    sub = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', string)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', sub).lower()

def _int_float_str(value):
    # Parse value type.
    try:
        return int(value)
    except ValueError:
        try:
            return float(value)
        except ValueError:
            return value

def _params(params):
    # Parse input parameters
    out = {}
    key = None
    for param in params:
        if param.startswith('--'):
            if '=' in param:
                key = param[2:].split('=')
                param = ','.join(key[1:])
                key = _underscore_case(key[0])
            else:
                key = _underscore_case(param[2:])
                continue

        if key is None:
            continue

        for value in _split(param):
            if value == '':
                continue

            value = _int_float_str(value)

            if key not in out:
                out[key] = value
            elif not isinstance(out[key], list):
                out[key] = [out[key], value]
            else:
                out[key] += [value]

    return out

def cli_calculation(argv=None, calculation=Calculation, desc='generic'):
    '''Submit and get calculation results.

    - POST: /calculation/new + payload
    - GET: /calculation/{id}
    - GET: /calculation/{id}/results
    '''
    parser = argparse.ArgumentParser(
        description=f'Submit {desc} calculation to the WebGeoCalc API')

    parser.add_argument('--quiet', '-q', action='store_true',
                        help='Disable verbose output status.')
    parser.add_argument('--payload', '-p', action='store_true',
                        help='Display payload before the calculation results.')
    parser.add_argument('--dry-run', '-d', action='store_true',
                        help='Dry run. Show only the payload.')
    parser.add_argument('--KEY', metavar='VALUE', nargs='*',
                        help='Key parameter and its value(s).')

    args, others = parser.parse_known_args(argv)
    params = _params(others)

    if params:
        params['verbose'] = not args.quiet
        try:
            calc = calculation(**params)

            if (args.payload or args.dry_run) and not args.quiet:

                payload = calc.payload
                print('Payload:\n{')

                for key, value in payload.items():
                    print(f'  {key}: {value},')
                print('}')

            if not args.dry_run:
                if not args.quiet:
                    print('\nAPI status:')

                res = calc.run()
                if not args.quiet:
                    print('\nResults:')

                for key, value in res.items():
                    print(f'{key}:\n> {value}')

        except (AttributeError, ValueError, IOError) as err:
            print(err)
    else:
        parser.print_help()


def cli_state_vector(argv=None):
    '''Submit state vector calculation with the CLI.'''
    cli_calculation(argv, StateVector, desc='State Vector')

def cli_angular_separation(argv=None):
    '''Submit angular separation calcultion with the CLI.'''
    cli_calculation(argv, AngularSeparation, desc='Angular Separation')

def cli_angular_size(argv=None):
    '''Submit angular size calcultion with the CLI.'''
    cli_calculation(argv, AngularSize, desc='Angular Size')

def cli_frame_transformation(argv=None):
    '''Submit frame transformation calcultion with the CLI.'''
    cli_calculation(argv, FrameTransformation, desc='Frame Transformation')

def cli_illumination_angles(argv=None):
    '''Submit illumination angles calcultion with the CLI.'''
    cli_calculation(argv, IlluminationAngles, desc='Illumination Angles')

def cli_subsolar_point(argv=None):
    '''Submit sub-solar point calcultion with the CLI.'''
    cli_calculation(argv, SubSolarPoint, desc='Sub-Solar Point')

def cli_subobserver_point(argv=None):
    '''Submit sub-observer point calcultion with the CLI.'''
    cli_calculation(argv, SubObserverPoint, desc='Sub-Observer Point')

def cli_surface_intercept_point(argv=None):
    '''Submit surface intercept point calcultion with the CLI.'''
    cli_calculation(argv, SurfaceInterceptPoint, desc='Surface Intercept Point')

def cli_osculating_elements(argv=None):
    '''Submit osculating elements calcultion with the CLI.'''
    cli_calculation(argv, OsculatingElements, desc='Osculating Elements')

def cli_time_conversion(argv=None):
    '''Submit time conversion calcultion with the CLI.'''
    cli_calculation(argv, TimeConversion, desc='Time Conversion')



# -*- coding: utf-8 -*-
'''Test WGC calculation runs.'''

import pytest

from webgeocalc import Calculation, StateVector
from webgeocalc.errors import (CalculationAlreadySubmitted, CalculationFailed,
                               CalculationNotCompleted, CalculationTimeOut,
                               ResultAttributeError)
from webgeocalc.vars import API_URL

@pytest.fixture
def params():
    '''Input parameters with 2 input time for generic calculation.'''
    return {
        "calculation_type": "STATE_VECTOR",
        "kernels": 1,
        "time_system": "UTC",
        "time_format": "CALENDAR",
        "intervals": ["2000-01-01", "2000-01-02"],
        "time_step": 30,
        "time_step_units": "MINUTES",
        "target": "SUN",
        "observer": "EARTH",
        "reference_frame": "J2000",
        "aberration_correction": "NONE",
        "state_representation": "RA_DEC"
    }

@pytest.fixture
def response():
    '''Response status expected from the API.'''
    return {
        "status": "OK",
        "message": "The operation was successful.",
        "calculationId": "0788aba2-d4e5-4028-9ef1-4867ad5385e0",
        "result": {
            "phase": "COMPLETE",
            "progress": 0
        }
    }

@pytest.fixture
def results():
    '''Results expected from the API.'''
    return {
        "status": "OK",
        "message": "The operation was successful.",
        "calculationId": "0788aba2-d4e5-4028-9ef1-4867ad5385e0",
        "columns": [
            {
                "name": "UTC calendar date",
                "type": "DATE",
                "outputID": "DATE",
                "units": ""
            },
            {
                "name": "Right Ascension (deg)",
                "type": "NUMBER",
                "outputID": "RIGHT_ASCENSION",
                "units": "deg"
            },
            {
                "name": "Declination (deg)",
                "type": "NUMBER",
                "outputID": "DECLINATION",
                "units": "deg"
            },
            {
                "name": "Range (km)",
                "type": "NUMBER",
                "outputID": "RANGE",
                "units": "km"
            },
            {
                "name": "d Right Ascension/dt (deg/s)",
                "type": "NUMBER",
                "outputID": "D_RIGHT_ASCENSION_DT",
                "units": "deg/s"
            },
            {
                "name": "d Declination/dt (deg/s)",
                "type": "NUMBER",
                "outputID": "D_DECLINATION_DT",
                "units": "deg/s"
            },
            {
                "name": "d Range/dt (km/s)",
                "type": "NUMBER",
                "outputID": "D_RANGE_DT",
                "units": "km/s"
            },
            {
                "name": "Speed (km/s)",
                "type": "NUMBER",
                "outputID": "SPEED",
                "units": "km/s"
            },
            {
                "name": "Time at Target",
                "type": "DATE",
                "outputID": "TIME_AT_TARGET",
                "units": ""
            },
            {
                "name": "Light Time (s)",
                "type": "NUMBER",
                "outputID": "LIGHT_TIME",
                "units": "s"
            }
        ],
        "rows": [
            [
                "2000-01-01 00:00:00.000000 UTC",
                280.73666816,
                -23.07197828,
                147104359.15044397,
                0.0000127884715,
                8.74255781e-7,
                -0.01659221,
                30.29084386,
                "2000-01-01 00:00:00.000000 UTC",
                490.6873246
            ],
            [
                "2000-01-02 00:00:00.000000 UTC",
                281.84100169,
                -22.99260787,
                147103258.34983575,
                0.0000127745341,
                9.62949471e-7,
                -0.00894244,
                30.29324424,
                "2000-01-02 00:00:00.000000 UTC",
                490.68365272
            ]
        ]
    }


@pytest.fixture
def params_sv():
    '''Input parameters for state vector caclulation (1 time only).'''
    return {
        "kernels": 5,
        "times": "2012-10-19T08:24:00.000",
        "target": "ENCELADUS",
        "observer": "CASSINI",
        "reference_frame": "CASSINI_ISS_NAC",
    }

@pytest.fixture
def response_sv():
    '''State vector response status expected from the API.'''
    return {
        "status": "OK",
        "message": "The operation was successful.",
        "calculationId": "5d009079-aa9e-4fbd-93c3-58b5a4990a68",
        "result": {
            "phase": "COMPLETE",
            "progress": 0
        }
    }

@pytest.fixture
def results_sv():
    '''State vector esults expected from the API.'''
    return {
        "status": "OK",
        "message": "The operation was successful.",
        "calculationId": "5d009079-aa9e-4fbd-93c3-58b5a4990a68",
        "columns": [
            {
                "name": "UTC calendar date",
                "type": "DATE",
                "outputID": "DATE",
                "units": ""
            },
            {
                "name": "Distance (km)",
                "type": "NUMBER",
                "outputID": "DISTANCE",
                "units": "km"
            },
            {
                "name": "Speed (km/s)",
                "type": "NUMBER",
                "outputID": "SPEED",
                "units": "km/s"
            },
            {
                "name": "X (km)",
                "type": "NUMBER",
                "outputID": "X",
                "units": "km"
            },
            {
                "name": "Y (km)",
                "type": "NUMBER",
                "outputID": "Y",
                "units": "km"
            },
            {
                "name": "Z (km)",
                "type": "NUMBER",
                "outputID": "Z",
                "units": "km"
            },
            {
                "name": "dX/dt (km/s)",
                "type": "NUMBER",
                "outputID": "D_X_DT",
                "units": "km/s"
            },
            {
                "name": "dY/dt (km/s)",
                "type": "NUMBER",
                "outputID": "D_Y_DT",
                "units": "km/s"
            },
            {
                "name": "dZ/dt (km/s)",
                "type": "NUMBER",
                "outputID": "D_Z_DT",
                "units": "km/s"
            },
            {
                "name": "Time at Target",
                "type": "DATE",
                "outputID": "TIME_AT_TARGET",
                "units": ""
            },
            {
                "name": "Light Time (s)",
                "type": "NUMBER",
                "outputID": "LIGHT_TIME",
                "units": "s"
            }
        ],
        "rows": [
            [
                "2012-10-19 08:24:00.000000 UTC",
                967899.52452788,
                12.31257026,
                -112.54927003,
                -316.50526854,
                967899.4662352,
                9.20424051,
                7.86483252,
                2.24181902,
                "2012-10-19 08:23:56.771435 UTC",
                3.22856529
            ]
        ]
    }

@pytest.fixture
def loading_kernels():
    '''Uncomplete response status expected from the API.'''
    return {
        "status": "OK",
        "message": "Loading kernels…",
        "calculationId": "824e983c-f75f-4f49-89bd-b487a77da65c",
        "result": {
            "phase": "LOADING_KERNELS",
            "progress": 0
        }
    }

def test_calculation_run(requests_mock, params, response, results):
    '''Run generic calculation.'''
    calc = Calculation(**params)

    requests_mock.post(API_URL + '/calculation/new', json=response)
    requests_mock.get(
        API_URL + '/calculation/' + response['calculationId'], json=response)
    requests_mock.get(
        API_URL + '/calculation/' + response['calculationId'] + '/results', json=results)

    with pytest.raises(CalculationNotCompleted):
        _ = calc.results

    calc.run()
    assert calc.id == response['calculationId']

    calc.update()
    assert calc.phase == response['result']['phase']

    with pytest.raises(CalculationAlreadySubmitted):
        calc.submit()

    calc.resubmit()
    assert calc.id == response['calculationId']

    out = calc.run()  # Re-run results without API request
    assert len(out['DATE']) == len(results['rows'])

def test_state_vector_single_time(requests_mock, params_sv, response_sv, results_sv):
    '''Run state vector calculation.'''
    sv = StateVector(**params_sv)

    requests_mock.post(API_URL + '/calculation/new', json=response_sv)
    requests_mock.get(
        API_URL + '/calculation/' + response_sv['calculationId'], json=response_sv)
    requests_mock.get(
        API_URL + '/calculation/' + response_sv['calculationId'] + '/results',
        json=results_sv)

    out = sv.run()
    assert out['DATE'] == results_sv['rows'][0][0]

    column = sv.columns[0]
    column_sv = results_sv['columns'][0]

    assert str(column) == column_sv['name']

    for key in column.keys():
        assert key in column_sv.keys()

    for value in column.values():
        assert value in column_sv.values()

    for key, value in column.items():
        assert key in column_sv.keys()
        assert value in column_sv.values()

    with pytest.raises(ResultAttributeError):
        _ = column.wrong_attr

def test_calculation_cancel(params):
    '''Test error if calculation is cancelled.'''
    calc = Calculation(**params)
    calc.submit()
    calc.cancel()

    assert calc.phase == 'CANCELLED'

    with pytest.raises(CalculationFailed):
        calc.run()

def test_calculation_timeout(requests_mock, params, loading_kernels):
    '''Test error if response exceed timeout.'''
    requests_mock.post(API_URL + '/calculation/new', json=loading_kernels)

    with pytest.raises(CalculationTimeOut):
        Calculation(**params).run(timeout=0.001, sleep=0.001)



# -*- coding: utf-8 -*-
'''Test WGC angular separation calculation.'''

import pytest

from webgeocalc import AngularSeparation

@pytest.fixture
def kernel_paths():
    '''Input kernel paths.'''
    return [
        'pds/wgc/kernels/lsk/naif0012.tls',
        'pds/wgc/kernels/spk/de430.bsp',
    ]

@pytest.fixture
def time():
    '''Input time.'''
    return '2012-10-19T08:24:00.000'

@pytest.fixture
def target_1():
    '''Input name for the first target.'''
    return 'VENUS'

@pytest.fixture
def target_2():
    '''Input name for the second target.'''
    return 'MERCURY'

@pytest.fixture
def observer():
    '''Input name of the observer.'''
    return 'SUN'

@pytest.fixture
def corr():
    '''Input aberration correction.'''
    return 'NONE'

@pytest.fixture
def params(kernel_paths, time, target_1, target_2, observer, corr):
    '''Input parameters from WGC API example.'''
    return {
        'kernel_paths': kernel_paths,
        'times': time,
        'target_1': target_1,
        'target_2': target_2,
        'observer': observer,
        'aberration_correction': corr,
    }

@pytest.fixture
def payload(kernel_paths, time, target_1, target_2, observer, corr):
    '''Payload from WGC API example.'''
    return {
        "kernels": [{
            "type": "KERNEL",
            "path": kernel_paths[0],
        }, {
            "type": "KERNEL",
            "path": kernel_paths[1],
        }],
        "timeSystem": "UTC",
        "timeFormat": "CALENDAR",
        "times": [
            time,
        ],
        "calculationType": "ANGULAR_SEPARATION",
        "target1": target_1,
        "shape1": "POINT",
        "target2": target_2,
        "shape2": "POINT",
        "observer": observer,
        "aberrationCorrection": corr
    }

def test_angular_separation_payload(params, payload):
    '''Test angular separation payload.'''
    assert AngularSeparation(**params).payload == payload



'''WebGeoCalc module.'''

import pbr.version

__version__ = pbr.version.VersionInfo('webgeocalc').version_string()

from .api import API
from .calculation import Calculation
from .calculation_types import (AngularSeparation, AngularSize,
                                FrameTransformation, IlluminationAngles,
                                OsculatingElements, StateVector,
                                SubObserverPoint, SubSolarPoint,
                                SurfaceInterceptPoint, TimeConversion)

__all__ = [
    'API',
    'AngularSeparation',
    'AngularSize',
    'Calculation',
    'FrameTransformation',
    'IlluminationAngles',
    'OsculatingElements',
    'StateVector',
    'SubObserverPoint',
    'SubSolarPoint',
    'SurfaceInterceptPoint',
    'TimeConversion',
    '__version__',
]



# -*- coding: utf-8 -*-
'''Test WGC command line inputs.'''

from webgeocalc.cli import (_params, cli_angular_separation, cli_angular_size, cli_bodies,
                            cli_frame_transformation, cli_frames, cli_illumination_angles,
                            cli_instruments, cli_kernel_sets, cli_osculating_elements,
                            cli_state_vector, cli_subobserver_point, cli_subsolar_point,
                            cli_surface_intercept_point, cli_time_conversion)

def test_cli_kernel_sets(capsys):
    '''Test GET kernels with CLI.'''
    argv = ''.split()
    cli_kernel_sets(argv)
    captured = capsys.readouterr()
    assert 'usage:' in captured.out

    argv = '--all'.split()
    cli_kernel_sets(argv)
    captured = capsys.readouterr()
    assert ' - Solar System Kernels: (id: 1)' in captured.out
    assert ' - Cassini Huygens: (id: 5)' in captured.out

    argv = '--kernel 1 Cassini'.split()
    cli_kernel_sets(argv)
    captured = capsys.readouterr()
    assert ' - Solar System Kernels: (id: 1)' in captured.out
    assert 'Too many kernel sets contains \'Cassini\' in their names:' in captured.out

def test_cli_bodies(capsys):
    '''Test GET bodies with CLI.'''
    argv = ''.split()
    cli_bodies(argv)
    captured = capsys.readouterr()
    assert 'usage:' in captured.out

    argv = ['Cassini Huygens']
    cli_bodies(argv)
    captured = capsys.readouterr()
    assert ' - CASSINI: (id: -82)' in captured.out
    assert ' - TITAN: (id: 606)' in captured.out

    argv = '5 --name Titan'.split()
    cli_bodies(argv)
    captured = capsys.readouterr()
    assert captured.out == ' - TITAN: (id: 606)\n'

def test_cli_frames(capsys):
    '''Test GET frames with CLI.'''
    argv = ''.split()
    cli_frames(argv)
    captured = capsys.readouterr()
    assert 'usage:' in captured.out

    argv = ['Cassini Huygens']
    cli_frames(argv)
    captured = capsys.readouterr()
    assert ' - IAU_TITAN: (id: 10044)' in captured.out
    assert ' - J2000: (id: 1)' in captured.out

    argv = '5 --name Titan'.split()
    cli_frames(argv)
    captured = capsys.readouterr()
    assert ' - IAU_TITAN: (id: 10044)' in captured.out
    assert ' - J2000: (id: 1)' not in captured.out

def test_cli_instruments(capsys):
    '''Test GET instruments with CLI.'''
    argv = ''.split()
    cli_instruments(argv)
    captured = capsys.readouterr()
    assert 'usage:' in captured.out

    argv = ['Cassini Huygens']
    cli_instruments(argv)
    captured = capsys.readouterr()
    assert ' - CASSINI_ISS_WAC: (id: -82361)' in captured.out
    assert ' - CASSINI_VIMS_IR: (id: -82370)' in captured.out

    argv = '5 --name ISS'.split()
    cli_instruments(argv)
    captured = capsys.readouterr()
    assert ' - CASSINI_ISS_WAC: (id: -82361)' in captured.out
    assert ' - CASSINI_VIMS_IR: (id: -82370)' not in captured.out

def test_cli_input_parameters():
    '''Test CLI input parameters parsing.'''
    def parse(x):
        return _params(x.split())

    assert parse('--kernels 1') == {'kernels': 1}
    assert parse('--kernels=1') == {'kernels': 1}
    assert parse('--kernels= 1') == {'kernels': 1}
    assert parse('--kernels = 1') == {'kernels': 1}
    assert parse('--kernels =1') == {'kernels': 1}
    assert parse('--kernels 1 5') == {'kernels': [1, 5]}
    assert parse('--kernels=1 5') == {'kernels': [1, 5]}
    assert parse('--kernels 1,5') == {'kernels': [1, 5]}
    assert parse('--kernels=1,5') == {'kernels': [1, 5]}
    assert parse('--kernels 1, 5') == {'kernels': [1, 5]}
    assert parse('--kernels=1, 5') == {'kernels': [1, 5]}
    assert parse('--kernels [1,5]') == {'kernels': [1, 5]}
    assert parse('--kernels=[1,5]') == {'kernels': [1, 5]}
    assert parse('--kernels [1, 5]') == {'kernels': [1, 5]}
    assert parse('--kernels=[1, 5]') == {'kernels': [1, 5]}
    assert parse('--kernels [ 1,5]') == {'kernels': [1, 5]}
    assert parse('--kernels=[ 1,5]') == {'kernels': [1, 5]}
    assert parse('--kernels [ 1, 5]') == {'kernels': [1, 5]}
    assert parse('--kernels [ 1, 5 ]') == {'kernels': [1, 5]}
    assert parse('--kernels [1, 5 ]') == {'kernels': [1, 5]}
    assert parse('--kernels [1,5 ]') == {'kernels': [1, 5]}
    assert parse('--kernels=[1,5 ]') == {'kernels': [1, 5]}
    assert parse('--kernels [1 ,5]') == {'kernels': [1, 5]}
    assert parse('--kernels [1 , 5]') == {'kernels': [1, 5]}
    assert parse('--kernels [1, 3, 5]') == {'kernels': [1, 3, 5]}
    assert parse('--kernels 1, 3, 5') == {'kernels': [1, 3, 5]}
    assert parse('--kernels 1 --kernels 5') == {'kernels': [1, 5]}
    assert parse('--kernels=1 --kernels 5') == {'kernels': [1, 5]}
    assert parse('--kernels=1 --kernels=5') == {'kernels': [1, 5]}
    assert parse('null --kernels=1') == {'kernels': 1}

    assert parse('--kernels Cassini') == {'kernels': 'Cassini'}
    assert parse("--kernels 'Cassini'") == {'kernels': 'Cassini'}
    assert parse('--kernels "Cassini"') == {'kernels': 'Cassini'}
    assert parse('--kernels "Cassini" "Solar"') == {'kernels': ['Cassini', 'Solar']}

    assert parse("--times '2012-10-19T08:24:00'") == {'times': '2012-10-19T08:24:00'}

def test_cli_state_vector_empty(capsys):
    '''Test empty state vector calculation parameter with the CLI.'''
    argv = ''.split()
    cli_state_vector(argv)
    captured = capsys.readouterr()
    assert 'usage:' in captured.out

def test_cli_angular_separation_wrong_attr(capsys):
    '''Test attribute in angular separation calculation parameter with the CLI.'''
    argv = '--wrong 123'.split()
    cli_angular_separation(argv)
    captured = capsys.readouterr()
    assert captured.out == 'Attribute \'target_1\' required.\n'

def test_cli_angular_size_run(capsys):
    '''Test run angular size calculation parameter with the CLI.'''
    argv = ('--payload '
            '--kernels 5 '
            '--times 2012-10-19T08:24:00 '
            '--target ENCELADUS '
            '--observer CASSINI '
            '--aberration_correction CN+S').split()

    cli_angular_size(argv)
    captured = capsys.readouterr()
    assert 'Payload:' in captured.out
    assert "kernels: [{'type': 'KERNEL_SET', 'id': 5}]," in captured.out
    assert "times: ['2012-10-19T08:24:00']," in captured.out
    assert "calculationType: ANGULAR_SIZE," in captured.out
    assert "timeSystem: UTC," in captured.out
    assert 'API status:\n[Calculation submit] Phase:' in captured.out
    assert 'Results:\nDATE:\n> 2012-10-19 08:24:00.000000 UTC' in captured.out
    assert 'ANGULAR_SIZE:\n> 0.03037939' in captured.out

def test_cli_frame_transformation_dry_run(capsys):
    '''Test dry-run frame transform calculation parameter with the CLI.'''
    argv = ('--dry-run '
            '--kernels 5 '
            '--times 2012-10-19T08:24:00.000 '
            '--frame_1 IAU_SATURN '
            '--frame_2 IAU_ENCELADUS '
            '--aberration_correction NONE').split()

    cli_frame_transformation(argv)
    captured = capsys.readouterr()
    assert 'Payload:' in captured.out
    assert "calculationType: FRAME_TRANSFORMATION," in captured.out
    assert 'API status:\n[Calculation submit] Status:' not in captured.out
    assert 'Results:' not in captured.out

def test_cli_illumination_angles_dry_run(capsys):
    '''Test dry-run illumination angles calculation parameter with the CLI.'''
    argv = ('--dry-run '
            '--kernels 5 '
            '--times 2012-10-19T08:24:00.000 '
            '--target ENCELADUS '
            '--target_frame IAU_ENCELADUS '
            '--observer CASSINI '
            '--aberration_correction CN+S '
            '--latitude 0.0 '
            '--longitude 0.0').split()

    cli_illumination_angles(argv)
    captured = capsys.readouterr()
    assert 'Payload:' in captured.out
    assert "calculationType: ILLUMINATION_ANGLES," in captured.out

def test_cli_subsolar_point_dry_run(capsys):
    '''Test dry-run sub-solar point calculation parameter with the CLI.'''
    argv = ('--dry-run '
            '--kernels 5 '
            '--times 2012-10-19T08:24:00.000 '
            '--target ENCELADUS '
            '--target_frame IAU_ENCELADUS '
            '--observer CASSINI '
            '--aberration_correction CN+S').split()

    cli_subsolar_point(argv)
    captured = capsys.readouterr()
    assert 'Payload:' in captured.out
    assert "calculationType: SUB_SOLAR_POINT," in captured.out

def test_cli_subobserver_point_dry_run(capsys):
    '''Test dry-run sub-observer point calculation parameter with the CLI.'''
    argv = ('--dry-run '
            '--kernels 5 '
            '--times 2012-10-19T08:24:00.000 '
            '--target ENCELADUS '
            '--target_frame IAU_ENCELADUS '
            '--observer CASSINI '
            '--aberration_correction CN+S').split()

    cli_subobserver_point(argv)
    captured = capsys.readouterr()
    assert 'Payload:' in captured.out
    assert "calculationType: SUB_OBSERVER_POINT," in captured.out

def test_cli_surface_intercept_point_dry_run(capsys):
    '''Test dry-run surface intercept point calculation parameter with the CLI.'''
    argv = ('--dry-run '
            '--kernels 5 '
            '--times 2012-10-19T08:24:00.000 '
            '--target SATURN '
            '--target_frame IAU_SATURN '
            '--observer CASSINI '
            '--direction_vector_type INSTRUMENT_BORESIGHT '
            '--direction_instrument CASSINI_ISS_NAC '
            '--aberration_correction NONE '
            '--state_representation LATITUDINAL').split()

    cli_surface_intercept_point(argv)
    captured = capsys.readouterr()
    assert 'Payload:' in captured.out
    assert "calculationType: SURFACE_INTERCEPT_POINT," in captured.out

def test_cli_osculating_elements_dry_run(capsys):
    '''Test dry-run osculating elements calculation parameter with the CLI.'''
    argv = ('--dry-run '
            '--kernels 1 5 '
            '--times 2012-10-19T08:24:00.000 '
            '--orbiting_body CASSINI '
            '--center_body SATURN').split()

    cli_osculating_elements(argv)
    captured = capsys.readouterr()
    assert 'Payload:' in captured.out
    assert "calculationType: OSCULATING_ELEMENTS," in captured.out

def test_cli_time_conversion_dry_run(capsys):
    '''Test dry-run time conversion calculation parameter with the CLI.'''
    argv = ('--dry-run '
            '--kernels 5 '
            '--times 1/1729329441.04 '
            '--time_system SPACECRAFT_CLOCK '
            '--time_format SPACECRAFT_CLOCK_STRING '
            '--sclk_id -82 '
            ).split()

    cli_time_conversion(argv)
    captured = capsys.readouterr()
    assert 'Payload:' in captured.out
    assert "calculationType: TIME_CONVERSION," in captured.out



# -*- coding: utf-8 -*-
'''WebGeoCalc API module.'''

import requests

from .errors import APIError, APIReponseError, KernelSetNotFound, TooManyKernelSets
from .types import ColumnResult, KernelSetDetails, get_type
from .vars import API_URL

class Api:
    '''WebGeoCalc API object.

    Parameters
    ----------
    url : str, optional
        API root URL. Default:
        ``https://wgc2.jpl.nasa.gov:8443/webgeocalc/api``

    '''

    def __init__(self, url=API_URL):
        self.url = url
        self._kernel_sets = None

    def get(self, url):
        '''Generic GET request on the API.

        Parameters
        ----------
        url: str
            Request url with GET method.

        Returns
        -------
        list or tuple:
            Parsed json API response. See: :py:func:`read`.

        Raises
        ------
        requests.reponse.HTMLError
            If HTML error is thrown by the API (HTML code not equal 200)

        Example
        -------
        >>> API.get('/kernel-sets')
        [<KernelSetDetails> Solar System Kernels (id: 1), ...]

        '''
        response = requests.get(self.url + url)
        if response.ok:
            return self.read(response.json())

        return response.raise_for_status()

    def post(self, url, payload):
        '''Generic POST request on the API.

        Parameters
        ----------
        url: str
            Request url with POST method.

        payload: dict
            Calculation payload data.

        Returns
        -------
        list or tuple:
            Parsed json API response. See: :py:func:`read`.

        Raises
        ------
        requests.reponse.HTMLError
            If HTML error is thrown by the API (HTML code not equal 200)

        Example
        -------
        >>> API.post('/calculation/new', payload=calculation_payload)  # doctest: +SKIP
        ('0788aba2-d4e5-4028-9ef1-4867ad5385e0', 'COMPLETE', 0)

        '''
        response = requests.post(self.url + url, json=payload)
        if response.ok:
            return self.read(response.json())

        return response.raise_for_status()

    @staticmethod
    def read(json):
        '''Read content from API JSON reponse.

        Parameters
        ----------
        json: dict
            API JSON response

        Returns
        -------
        list or tuple
            Parsed content of the JSON API response.

            If the response contents ``resultType`` and ``items`` keys, then
            return type will be:

            [`Result objects`: :obj:`webgeocalc.types.ResultType`]

            If response contents ``result``, then return type will be:

            (`Calculation id`: str, `Phase`: str)

            If the response contents ``columns`` and ``rows``, then
            return type will be:

            (`Columns`: [:obj:`webgeocalc.types.ColumnResult`],
             `Results rows`: [[int, float or str], ...])

            Else an error is thrown.

        Raises
        ------
        APIError
            If the API status is not ``OK``.
        APIReponseError
            If the format of the API response is unexpected.

        '''
        if json['status'] != 'OK':
            raise APIError(json['error']['shortDescription'])

        keys = json.keys()

        if 'resultType' in keys and 'items' in json:
            dtype = get_type(json['resultType'])
            return [dtype(item) for item in json['items']]

        if 'result' in keys:
            phase = json['result']['phase']
            if phase == 'QUEUED':
                phase += f' | POSITION: {json["result"]["position"]}'

            return json['calculationId'], phase

        if 'columns' in keys and 'rows' in keys:
            cols = [ColumnResult(col) for col in json['columns']]
            return cols, json['rows']

        raise APIReponseError(json)

    def kernel_sets(self):
        '''Get list of all kernel sets available on the webgeocalc server.

        ``GET: /kernel-sets``

        Returns
        -------
        [:obj:`webgeocalc.types.KernelSetDetails`]
            List of kernel sets.

        '''
        if self._kernel_sets is None:
            self._kernel_sets = self.get('/kernel-sets')
        return self._kernel_sets

    def kernel_set(self, kernel_set):
        '''Get kernel set by ``caption`` name or kernel set ``id``.

        Parameters
        ----------
        kernel_set: str or int
            Kernel set ``name`` or ``id``.

        Returns
        --------
        :obj:`webgeocalc.types.KernelSetDetails`
            Kernel set details object.

        '''
        kernel_sets = list(filter(lambda x: kernel_set in x, self.kernel_sets()))
        if len(kernel_sets) == 1:
            return kernel_sets[0]

        if len(kernel_sets) > 1:
            raise TooManyKernelSets(kernel_set, kernel_sets)

        raise KernelSetNotFound(kernel_set)

    def kernel_set_id(self, kernel_set):
        '''Extract kernel set ``id`` based on ``id``, ``name`` or `object`.

        Parameters
        ----------
        kernel_set: str, int or :obj:`webgeocalc.types.KernelSetDetails`
            Kernel sets ``name``, ``id`` or `object`.

        Returns
        -------
        int
            Kernel set ``id``.

        '''
        if isinstance(kernel_set, int):
            return kernel_set

        if isinstance(kernel_set, str):
            return int(self.kernel_set(kernel_set))

        if isinstance(kernel_set, KernelSetDetails):
            return int(kernel_set)

        raise TypeError(f"'kernel_set' must be a 'int', a 'str' of a 'KernelSetDetails' object:\n' + \
                        '>>> Type({kernel_set}) = {type(kernel_set)}")

    def bodies(self, kernel_set):
        '''Get list of bodies available in a kernel set.

        ``GET: /kernel-set/{kernelSetId}/bodies``

        Parameters
        ----------
        kernel_set: str, int or :obj:`webgeocalc.types.KernelSetDetails`
            Kernel sets ``name``, ``id`` or `object`.

        Returns
        -------
        [:obj:`webgeocalc.types.BodyData`]
            List of bodies in the requested kernel set.

        '''
        kernel_set_id = self.kernel_set_id(kernel_set)
        return self.get(f'/kernel-set/{kernel_set_id}/bodies')

    def frames(self, kernel_set):
        '''Get list of frames available in a kernel set.

        ``GET: /kernel-set/{kernelSetId}/frames``

        Parameters
        ----------
        kernel_set: str, int or :obj:`webgeocalc.types.KernelSetDetails`
            Kernel sets ``name``, ``id`` or `object`.

        Returns
        -------
        [:obj:`webgeocalc.types.FrameData`]
            List of frames in the requested kernel set.

        '''
        kernel_set_id = self.kernel_set_id(kernel_set)
        return self.get(f'/kernel-set/{kernel_set_id}/frames')

    def instruments(self, kernel_set):
        '''Get list of instruments available in a kernel set.

        ``GET: /kernel-set/{kernelSetId}/instruments``

        Parameters
        ----------
        kernel_set: str, int or :obj:`webgeocalc.types.KernelSetDetails`
            Kernel sets ``name``, ``id`` or `object`.

        Returns
        -------
        [:obj:`webgeocalc.types.InstrumentData`]
            List of instruments in the requested kernel set.

        '''
        kernel_set_id = self.kernel_set_id(kernel_set)
        return self.get(f'/kernel-set/{kernel_set_id}/instruments')

    def new_calculation(self, payload):
        '''Starts a new calculation.

        ``POST: /calculation/new``

        Parameters
        ----------
        payload: dict
            Calculation payload.

        Returns
        -------
        (str, str, int)
            Tuple of the calculation phase:
            ``(calculation-id, phase)``
            See: :py:func:`read`.

        Example
        -------
        >>> API.calculation_new(calculation_payload)  # doctest: +SKIP
        ('0788aba2-d4e5-4028-9ef1-4867ad5385e0', 'LOADING_KERNELS')

        '''
        return self.post('/calculation/new', payload)

    def phase_calculation(self, calculation_id):
        '''Gets the phase of a calculation.

        ``GET: /calculation/{id}``

        Parameters
        ----------
        calculation_id: str
            Calculation id.

        Returns
        -------
        (str, str)
            Tuple of the calculation phase:
            ``(calculation-id, phase)``
            See: :py:func:`read`.

        Example
        -------
        >>> API.phase_calculation('0788aba2-d4e5-4028-9ef1-4867ad5385e0')  # noqa: E501  # doctest: +SKIP
        ('0788aba2-d4e5-4028-9ef1-4867ad5385e0', 'COMPLETE')

        '''
        return self.get(f'/calculation/{calculation_id}')

    def cancel_calculation(self, calculation_id):
        '''Cancels a previously requested calculation, should this not be completed, or its results..

        ``GET: /calculation/{id}/cancel``

        Calculations that have been already ``DISPATCHED``, previously ``CANCELLED`` or
        that have ``EXPIRED`` cannot be cancelled, and requesting it will produce an `error`.

        Parameters
        ----------
        calculation_id: str
            Calculation id.

        Returns
        -------
        (str, str)
            Tuple of the calculation phase:
            ``(calculation-id, phase)``
            See: :py:func:`read`.

        Example
        -------
        >>> API.phase_calculation('0788aba2-d4e5-4028-9ef1-4867ad5385e0')  # noqa: E501  # doctest: +SKIP
        ('0788aba2-d4e5-4028-9ef1-4867ad5385e0', 'CANCELLED')

        '''
        return self.get(f'/calculation/{calculation_id}/cancel')

    def results_calculation(self, calculation_id):
        '''Gets the results of a complete calculation.

        ``GET: /calculation/{id}/results``

        Parameters
        ----------
        calculation_id: str
            Calculation id.

        Returns
        -------
        ([:obj:`webgeocalc.types.ColumnResult`], [[int, float or str], ...])
            Tuple of calculation results:
            ``(columns, rows)``
            See: :py:func:`read`.

        Example
        -------
        >>> API.results_calculation('0788aba2-d4e5-4028-9ef1-4867ad5385e0')  # noqa: E501  # doctest: +SKIP
        ([<ColumnResult> UTC calendar date, ...], [['2000-01-01 00:00:00.000000 UTC', ...], [...]])

        '''
        return self.get(f'/calculation/{calculation_id}/results')


# Export default API object
API = Api()



# -*- coding: utf-8 -*-
'''Test WGC illumination angles calculation.'''

import pytest

from webgeocalc import IlluminationAngles
from webgeocalc.errors import CalculationInvalidAttr, CalculationInvalidValue

@pytest.fixture
def kernels():
    '''Cassini kernel set.'''
    return 5

@pytest.fixture
def time():
    '''Input time.'''
    return '2012-10-19T08:24:00.000'

@pytest.fixture
def target():
    '''Input target.'''
    return 'ENCELADUS'

@pytest.fixture
def target_frame():
    '''Input frame.'''
    return 'IAU_ENCELADUS'

@pytest.fixture
def observer():
    '''Input oberserver.'''
    return 'CASSINI'

@pytest.fixture
def lat():
    '''Input latitude.'''
    return 0.0

@pytest.fixture
def lon():
    '''Input longitude.'''
    return 0.0

@pytest.fixture
def corr():
    '''Input aberration correction.'''
    return 'CN+S'

@pytest.fixture
def params(kernels, time, target, target_frame, observer, lat, lon, corr):
    '''Input parameters from WGC API example.'''
    return {
        'kernels': kernels,
        'times': time,
        'target': target,
        'target_frame': target_frame,
        'observer': observer,
        'latitude': lat,
        'longitude': lon,
        'aberration_correction': corr,
    }

@pytest.fixture
def payload(kernels, time, target, target_frame, observer, lat, lon, corr):
    '''Payload from WGC API example.'''
    return {
        "kernels": [{
            "type": "KERNEL_SET",
            "id": kernels,
        }],
        "timeSystem": "UTC",
        "timeFormat": "CALENDAR",
        "times": [
            time,
        ],
        "calculationType": "ILLUMINATION_ANGLES",
        "target": target,
        "shape1": "ELLIPSOID",
        "targetFrame": target_frame,
        "observer": observer,
        "coordinateRepresentation": "LATITUDINAL",
        "latitude": lat,
        "longitude": lon,
        "aberrationCorrection": corr,
    }

def test_illumination_angles_payload(params, payload):
    '''Test illumination angles payload.'''
    assert IlluminationAngles(**params).payload == payload

def test_illumination_angles_attr_error(params):
    '''Test erros whent illumination angles attribues are invalid.'''
    with pytest.raises(CalculationInvalidAttr):
        IlluminationAngles(shape_1='POINT', **params)

    with pytest.raises(CalculationInvalidAttr):
        IlluminationAngles(coordinate_representation='WRONG', **params)

def test_illumination_angles_value_error(params):
    '''Test latitude and longitude invalid inputs.'''
    params['latitude'] = 100
    with pytest.raises(CalculationInvalidValue):
        IlluminationAngles(**params)

    params['latitude'] = 90
    params['longitude'] = 190
    with pytest.raises(CalculationInvalidValue):
        IlluminationAngles(**params)



# -*- coding: utf-8 -*-
'''WebGeoCalc varibales.'''

API_URL = 'https://wgc2.jpl.nasa.gov:8443/webgeocalc/api'

CALCULATION_FAILED_PHASES = [
    'FAILED',
    'CANCELLED',
    'DISPATCHED',
    'EXPIRED',
]

CALCULATION_TYPE = [
    'STATE_VECTOR',
    'ANGULAR_SEPARATION',
    'ANGULAR_SIZE',
    'SUB_OBSERVER_POINT',
    'SUB_SOLAR_POINT',
    'ILLUMINATION_ANGLES',
    'SURFACE_INTERCEPT_POINT',
    'OSCULATING_ELEMENTS',
    'FRAME_TRANSFORMATION',
    'GF_COORDINATE_SEARCH',
    'GF_ANGULAR_SEPARATION_SEARCH',
    'GF_DISTANCE_SEARCH',
    'GF_SUB_POINT_SEARCH',
    'GF_OCCULTATION_SEARCH',
    'GF_TARGET_IN_INSTRUMENT_FOV_SEARCH',
    'GF_SURFACE_INTERCEPT_POINT_SEARCH',
    'GF_RAY_IN_FOV_SEARCH',
    'TIME_CONVERSION',
]

TIME_SYSTEM = [
    'UTC',
    'TDB',
    'TDT',
    'SPACECRAFT_CLOCK',
]

TIME_FORMAT = [
    'CALENDAR',
    'JULIAN',
    'SECONDS_PAST_J2000',
    'SPACECRAFT_CLOCK_TICKS',
    'SPACECRAFT_CLOCK_STRING',
]

OUTPUT_TIME_FORMAT = [
    'CALENDAR',
    'CALENDAR_YMD',
    'CALENDAR_DOY',
    'JULIAN',
    'SECONDS_PAST_J2000',
    'SPACECRAFT_CLOCK_STRING',
    'SPACECRAFT_CLOCK_TICKS',
    'CUSTOM',
]

TIME_STEP_UNITS = [
    'SECONDS',
    'MINUTES',
    'HOURS',
    'DAYS',
    'EQUAL_INTERVALS',
]

INTERVALS = [
    "array([startTime, endTime])",
    "dict({'startTime': startTime, 'endTime': endTime})",
    "array([ interval1, interval2, … ])",
]

ABERRATION_CORRECTION = [
    'NONE',
    'LT',
    'LT+S',
    'CN',
    'CN+S',
    'XLT',
    'XLT+S',
    'XCN',
    'XCN+S',
]

STATE_REPRESENTATION = [
    'RECTANGULAR',
    'RA_DEC',
    'LATITUDINAL',
    'PLANETODETIC',
    'PLANETOGRAPHIC',
    'CYLINDRICAL',
    'SPHERICAL',
]

SHAPE = [
    'POINT',
    'SPHERE',
    'ELLIPSOID',
    'DSK ',
]

TIME_LOCATION = [
    'FRAME1',
    'FRAME2',
]

ORIENTATION_REPRESENTATION = [
    'EULER_ANGLES',
    'ANGLE_AND_AXIS',
    'SPICE_QUATERNION',
    'OTHER_QUATERNION',
    'MATRIX_ROW_BY_ROW',
    'MATRIX_FLAGGED',
    'MATRIX_ALL_ONE_ROW',
]

ANGULAR_VELOCITY_REPRESENTATION = [
    'NOT_INCLUDED',
    'VECTOR_IN_FRAME1',
    'VECTOR_IN_FRAME2',
    'EULER_ANGLE_DERIVATIVES',
    'MATRIX',
]

AXIS = [
    'X',
    'Y',
    'Z',
]

ANGULAR_UNITS = [
    'deg',
    'rad',
]

ANGULAR_VELOCITY_UNITS = [
    'deg/s',
    'rad/s',
    'RPM',
    'Unitary',
]

COORDINATE_REPRESENTATION = [
    'LATITUDINAL',
    'PLANETODETIC',
    'PLANETOGRAPHIC',
]

SUB_POINT_TYPE = [
    'Near point: ellipsoid',
    'Intercept: ellipsoid',
    'NADIR/DSK/UNPRIORITIZED',
    'INTERCEPT/DSK/UNPRIORITIZED',
]

DIRECTION_VECTOR_TYPE = [
    'INSTRUMENT_BORESIGHT',
    'INSTRUMENT_FOV_BOUNDARY_VECTORS',
    'REFERENCE_FRAME_AXIS',
    'VECTOR_IN_INSTRUMENT_FOV',
    'VECTOR_IN_REFERENCE_FRAME',
]



# -*- coding: utf-8 -*-
'''Test WGC angular frame transformation calculation.'''

import pytest

from webgeocalc import FrameTransformation
from webgeocalc.errors import CalculationIncompatibleAttr, CalculationInvalidAttr

@pytest.fixture
def kernels():
    '''Cassini kernel set.'''
    return 5

@pytest.fixture
def time():
    '''Input time.'''
    return '2012-10-19T08:24:00.000'

@pytest.fixture
def frame_1():
    '''Input frame 1.'''
    return 'IAU_SATURN'

@pytest.fixture
def frame_2():
    '''Input frame 2.'''
    return 'IAU_ENCELADUS'

@pytest.fixture
def corr():
    '''Input aberration correction.'''
    return 'NONE'

@pytest.fixture
def params(kernels, time, frame_1, frame_2, corr):
    '''Input parameters from WGC API example.'''
    return {
        'kernels': kernels,
        'times': time,
        'frame_1': frame_1,
        'frame_2': frame_2,
        'aberration_correction': corr,
    }

@pytest.fixture
def payload(kernels, time, frame_1, frame_2, corr):
    '''Payload from WGC API example.'''
    return {
        "kernels": [{
            "type": "KERNEL_SET",
            "id": kernels,
        }],
        "timeSystem": "UTC",
        "timeFormat": "CALENDAR",
        "times": [
            time,
        ],
        "calculationType": "FRAME_TRANSFORMATION",
        "frame1": frame_1,
        "frame2": frame_2,
        "aberrationCorrection": corr,
        "timeLocation": "FRAME1",
        "orientationRepresentation": "EULER_ANGLES",
        "axis1": "X",
        "axis2": "Y",
        "axis3": "Z",
        "angularUnits": "deg",
        "angularVelocityRepresentation": "VECTOR_IN_FRAME1",
        "angularVelocityUnits": "deg/s"
    }

def test_frame_transformation_payload(params, payload):
    '''Test angular frame transformation payload.'''
    assert FrameTransformation(**params).payload == payload

def test_frame_transformation_attr_err(params):
    '''Test errors when frame transformation is invalid.'''
    del params['aberration_correction']
    with pytest.raises(CalculationInvalidAttr):
        # aberration_correctin can not be '+S'
        FrameTransformation(aberration_correction='CN+S', **params)

    with pytest.raises(CalculationInvalidAttr):
        FrameTransformation(time_location='WRONG', **params)

    with pytest.raises(CalculationInvalidAttr):
        FrameTransformation(orientation_representation='WRONG', **params)

    with pytest.raises(CalculationInvalidAttr):
        FrameTransformation(axis_1='WRONG', **params)

    with pytest.raises(CalculationInvalidAttr):
        FrameTransformation(angular_units='WRONG', **params)

    with pytest.raises(CalculationInvalidAttr):
        FrameTransformation(angular_velocity_representation='WRONG', **params)

    with pytest.raises(CalculationInvalidAttr):
        FrameTransformation(angular_velocity_units='WRONG', **params)

    with pytest.raises(CalculationIncompatibleAttr):
        FrameTransformation(angular_velocity_representation='EULER_ANGLE_DERIVATIVES',
                            angular_velocity_units='Unitary', **params)



# -*- coding: utf-8 -*-
'''Webgeocalc Calculations.'''

import time

from .api import API
from .errors import (CalculationAlreadySubmitted, CalculationConflictAttr,
                     CalculationFailed, CalculationIncompatibleAttr,
                     CalculationInvalidAttr, CalculationInvalidValue,
                     CalculationNotCompleted, CalculationRequiredAttr,
                     CalculationTimeOut, CalculationUndefinedAttr)
from .types import KernelSetDetails
from .vars import (ABERRATION_CORRECTION, ANGULAR_UNITS, ANGULAR_VELOCITY_REPRESENTATION,
                   ANGULAR_VELOCITY_UNITS, AXIS, CALCULATION_FAILED_PHASES,
                   CALCULATION_TYPE, COORDINATE_REPRESENTATION, DIRECTION_VECTOR_TYPE,
                   INTERVALS, ORIENTATION_REPRESENTATION, OUTPUT_TIME_FORMAT, SHAPE,
                   STATE_REPRESENTATION, SUB_POINT_TYPE, TIME_FORMAT, TIME_LOCATION,
                   TIME_STEP_UNITS, TIME_SYSTEM)


class SetterProperty:
    '''Setter property decorator.'''

    def __init__(self, func, doc=None):
        self.func = func
        self.__doc__ = doc if doc is not None else func.__doc__

    def __set__(self, obj, value):
        return self.func(obj, value)


class Calculation:
    '''Webgeocalc calculation object.

    Parameters
    ----------
    time_system: str, optional
        See: :py:attr:`time_system`
    time_format: str, optional
        See: :py:attr:`time_format`
    verbose: bool, optional
        Verbose calculation phase during :py:func:`submit`, :py:func:`update`
        and :py:func:`run`.

    Other Parameters
    ----------------
    calculation_type: str
        See: :py:attr:`calculation_type`
    kernels: str, int, [str or/and int]
        See: :py:attr:`kernels`
    kernel_paths: str, [str]
        See: :py:attr:`kernel_paths`
    times: str or [str]
        See: :py:attr:`times`
    intervals: [str, str] or {'startTime': str, 'endTime': str} or [interval, ...]
        See: :py:attr:`intervals`
    time_step: int
        See: :py:attr:`time_step`
    time_step_units: str
        See: :py:attr:`time_step_units`
    sclk_id: int
        See: :py:attr:`sclk_id`
    output_time_system: str
        See: :py:attr:`output_time_system`
    output_time_format: str
        See: :py:attr:`output_time_format`
    output_time_custom_format: str
        See: :py:attr:`output_time_custom_format`
    output_sclk_id: int
        See: :py:attr:`output_sclk_id`
    target: str or int
        See: :py:attr:`target`
    target_frame: str
        See: :py:attr:`target_frame`
    target_1: str
        See: :py:attr:`target_1`
    target_2: str
        See: :py:attr:`target_2`
    shape_1: str
        See: :py:attr:`shape_1`
    shape_2: str
        See: :py:attr:`shape_2`
    observer: str or int
        See: :py:attr:`observer`
    reference_frame: str or int
        See: :py:attr:`reference_frame`
    frame_1: str ot int
        See: :py:attr:`frame_1`
    frame_2: str or int
        See: :py:attr:`frame_2`
    orbiting_body: str
        See: :py:attr:`orbiting_body`
    center_body: str
        See: :py:attr:`center_body`
    aberration_correction: str
        See: :py:attr:`aberration_correction`
    state_representation: str
        See: :py:attr:`state_representation`
    time_location: str
        See: :py:attr:`time_location`
    orientation_representation: str
        See: :py:attr:`orientation_representation`
    axis_1: str
        See: :py:attr:`axis_1`
    axis_2: str
        See: :py:attr:`axis_2`
    axis_3: str
        See: :py:attr:`axis_3`
    angular_units: str
        See: :py:attr:`angular_units`
    angular_velocity_representation: str
        See: :py:attr:`angular_velocity_representation`
    angular_velocity_units: str
        See: :py:attr:`angular_velocity_units`
    coordinate_representation: str
        See: :py:attr:`coordinate_representation`
    latitude: float
        See: :py:attr:`latitude`
    longitude: float
        See: :py:attr:`longitude`
    sub_point_type: str
        See: :py:attr:`sub_point_type`
    direction_vector_type: str
        See: :py:attr:`direction_vector_type`
    direction_instrument: str or int
        See: :py:attr:`direction_instrument`
    direction_frame: str
        See: :py:attr:`direction_frame`
    direction_frame_axis: str
        See: :py:attr:`direction_frame_axis`
    direction_vector_x: float
        See: :py:attr:`direction_vector_x`
    direction_vector_y: float
        See: :py:attr:`direction_vector_y`
    direction_vector_z: float
        See: :py:attr:`direction_vector_z`
    direction_vector_ra: float
        See: :py:attr:`direction_vector_ra`
    direction_vector_dec: float
        See: :py:attr:`direction_vector_dec`

    Raises
    ------
    CalculationRequiredAttr
        If :py:attr:`calculation_type`, :py:attr:`time_system` and
        :py:attr:`time_format` are not provided.
    CalculationRequiredAttr
        If neither :py:attr:`kernels` nor :py:attr:`kernel_paths` is not provided.
    CalculationRequiredAttr
        If neither :py:attr:`times` nor :py:attr:`intervals` is not provided.

    '''

    def __init__(self, time_system='UTC', time_format='CALENDAR', verbose=True, **kwargs):
        # Add default parameters to kwargs
        kwargs['time_system'] = time_system
        kwargs['time_format'] = time_format

        # Init parameters
        self.params = kwargs
        self.__kernels = []
        self.id = None
        self.phase = 'NOT SUBMITED'
        self.columns = None
        self.values = None
        self.verbose = verbose

        # Required parameters
        self._required(['calculation_type', 'time_system', 'time_format'], kwargs)

        if 'kernels' not in kwargs.keys() and 'kernel_paths' not in kwargs.keys():
            raise CalculationRequiredAttr('kernels\' or \'kernel_paths')

        if 'times' not in kwargs.keys() and 'intervals' not in kwargs.keys():
            raise CalculationRequiredAttr('times\' or \'intervals')

        # Check and set parameters
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __repr__(self):
        return '\n'.join([
            f"<{self.__class__.__name__}> Phase: {self.phase} (id: {self.id})"
        ] + [
            f' - {k}: {v}' for k, v in self.payload.items()
        ])

    @staticmethod
    def _required(attrs, kwargs):
        # Check required arguments
        for required in attrs:
            if required not in kwargs.keys():
                raise CalculationRequiredAttr(required)

    @property
    def payload(self):
        '''Calculation payload parameters *dict* for JSON input in WebGeoCalc format.

        Return
        ------
        dict
            Payload keys and values.

        Example
        -------
        >>> Calculation(
        ...    kernels = 'Cassini Huygens',
        ...    times = '2012-10-19T08:24:00.000',
        ...    calculation_type = 'STATE_VECTOR',
        ...    target = 'CASSINI',
        ...    observer = 'SATURN',
        ...    reference_frame = 'IAU_SATURN',
        ...    aberration_correction = 'NONE',
        ...    state_representation = 'PLANETOGRAPHIC',
        ... ).payload  # noqa: E501
        {'kernels': [{'type': 'KERNEL_SET', 'id': 5}], 'times': ['2012-10-19T08:24:00.000'], ...}

        '''
        return {k.split('__')[-1]: v for k, v in vars(self).items() if k.startswith('_')}

    def submit(self):
        '''Submit calculation parameters and get calculation ``id`` and ``phase``.

        Raises
        ------
        CalculationAlreadySubmitted
            If the calculation was already submitted (to avoid duplicated submissions).

        Example
        -------
        >>> calc.submit()  # noqa: E501  # doctest: +SKIP
        [Calculation submit] Phase: QUEUED | POSITION: 6 (id: 8750344d-645d-4e43-b159-c8d88d28aac6)
        >>> calc.id        # doctest: +SKIP
        '8750344d-645d-4e43-b159-c8d88d28aac6'
        >>> calc.phase     # doctest: +SKIP
        'QUEUED | POSITION: 6'

        '''
        if self.id is not None:
            raise CalculationAlreadySubmitted(self.id)

        self.id, self.phase = API.new_calculation(self.payload)

        if self.verbose:
            print(f'[Calculation submit] Phase: {self.phase} (id: {self.id})')

    def resubmit(self):
        '''Reset calculation ``id`` and re-submit the calculation.

        See: :py:func:`submit`.
        '''
        self.id = None
        self.submit()

    def cancel(self):
        '''Cancels calculation if already submitted.'''
        if self.id is not None:
            _, self.phase = API.cancel_calculation(self.id)

            if self.verbose:
                print(f'[Calculation cancellation] Phase: {self.phase} (id: {self.id})')

    def update(self):
        '''Update calculation phase ``phase``.

        Example
        -------
        >>> calc.update()  # noqa: E501  # doctest: +SKIP
        [Calculation update] Phase: QUEUED | POSITION: 3 (id: 8750344d-645d-4e43-b159-c8d88d28aac6)
        >>> calc.update()  # noqa: E501  # doctest: +SKIP
        [Calculation update] Phase: STARTING (id: 8750344d-645d-4e43-b159-c8d88d28aac6)
        >>> calc.update()  # noqa: E501  # doctest: +SKIP
        [Calculation update] Phase: LOADING_KERNELS (id: 8750344d-645d-4e43-b159-c8d88d28aac6)
        >>> calc.update()  # noqa: E501  # doctest: +SKIP
        [Calculation update] Phase: CALCULATING (id: 8750344d-645d-4e43-b159-c8d88d28aac6)
        >>> calc.update()  # doctest: +SKIP
        [Calculation update] Phase: COMPLETE (id: 8750344d-645d-4e43-b159-c8d88d28aac6)
        '''
        if self.id is None:
            self.submit()
        else:
            _, self.phase = API.phase_calculation(self.id)

            if self.verbose:
                print(f'[Calculation update] Phase: {self.phase} (id: {self.id})')

    @property
    def results(self):
        '''Gets the results of a calculation, if its phase is `COMPLETE`.

        Return
        ------
        dict
            Calculation results as *dict* based on output columns. If multiple ``times``
            or ``intervals`` are used, the value of the *dict* will be an array.
            See examples.

        Raises
        ------
        CalculationNotCompleted
            If calculation phase is not `COMPLETE`.

        Examples
        --------
        >>> calc.results     # doctest: +SKIP
        {'DATE': '2012-10-19 09:00:00.000000 UTC',
         'DISTANCE': 764142.63776247,
         'SPEED': 111.54765899,
         'X': 298292.85744169,
         'Y': -651606.58468976,
         'Z': 265224.81187627,
         'D_X_DT': -98.8032491,
         'D_Y_DT': -51.73211296,
         'D_Z_DT': -2.1416539,
         'TIME_AT_TARGET': '2012-10-19 08:59:57.451094 UTC',
         'LIGHT_TIME': 2.54890548}

        >>> ang_sep = AngularSeparation(
        ...     kernel_paths = ['pds/wgc/kernels/lsk/naif0012.tls', 'pds/wgc/kernels/spk/de430.bsp'],
        ...     times = ['2012-10-19T08:24:00.000', '2012-10-19T09:00:00.000'],
        ...     target_1 = 'VENUS',
        ...     target_2 = 'MERCURY',
        ...     observer = 'SUN',
        ...     verbose = False,
        ... )  # noqa: E501
        >>> ang_sep.submit()      # doctest: +SKIP
        >>> ang_sep.results       # doctest: +SKIP
        {'DATE': ['2012-10-19 08:24:00.000000 UTC', '2012-10-19 09:00:00.000000 UTC'],
         'ANGULAR_SEPARATION': [175.17072258, 175.18555938]}

        '''
        if self.phase != 'COMPLETE':
            raise CalculationNotCompleted(self.phase)

        if self.columns is None or self.values is None:
            self.columns, self.values = API.results_calculation(self.id)

        if len(self.values) == 1:
            data = self.values[0]
        else:
            # Transpose values array
            data = [[row[i] for row in self.values] for i in range(len(self.columns))]

        return {column.outputID: value for column, value in zip(self.columns, data)}

    def run(self, timeout=30, sleep=1):
        '''Submit, update and retrive calculation results at once.

        See: :py:func:`submit`, :py:func:`update` and :py:attr:`results`.

        Parameters
        ----------
        timeout: int, optional
            Auto-update time out (in seconds).
        sleep: int, optional
            Sleep duration (in seconds) between each update.

        Raises
        ------
        CalculationTimeOut
            If calculation reach the timeout duration.

        '''
        if self.columns is not None and self.values is not None:
            return self.results

        for i in range(int(timeout / sleep)):
            self.update()

            if self.phase == 'COMPLETE':
                return self.results

            if self.phase in CALCULATION_FAILED_PHASES:
                raise CalculationFailed(self.phase)

            time.sleep(sleep)

        raise CalculationTimeOut(timeout, sleep)

    @SetterProperty
    def calculation_type(self, val):
        '''The type of calculation to perform.

        Parameters
        ----------
        calculation_type: str
            One of the following:

            - STATE_VECTOR
            - ANGULAR_SEPARATION
            - ANGULAR_SIZE
            - SUB_OBSERVER_POINT
            - SUB_SOLAR_POINT
            - ILLUMINATION_ANGLES
            - SURFACE_INTERCEPT_POINT
            - OSCULATING_ELEMENTS
            - FRAME_TRANSFORMATION
            - TIME_CONVERSION
            - GF_COORDINATE_SEARCH
            - GF_ANGULAR_SEPARATION_SEARCH
            - GF_DISTANCE_SEARCH
            - GF_SUB_POINT_SEARCH
            - GF_OCCULTATION_SEARCH
            - GF_TARGET_IN_INSTRUMENT_FOV_SEARCH
            - GF_SURFACE_INTERCEPT_POINT_SEARCH
            - GF_RAY_IN_FOV_SEARCH

            Note
            ----
            This parameters will be auto-filled for specific calculation sub-classes.

        Raises
        ------
        CalculationInvalidAttr
            If the value provided is invalid.

        '''
        if val in CALCULATION_TYPE:
            self.__calculationType = val
        else:
            raise CalculationInvalidAttr('calculation_type', val, CALCULATION_TYPE)

    @SetterProperty
    def kernels(self, kernel_sets):
        '''Add kernel sets.

        Parameters
        ----------
        kernels: str, int, [str or/and int]
            Kernel set(s) to be used for the calculation::

               [{'type': 'KERNEL_SET', 'id': 5}, ...]

        '''
        self.__kernels += [
            self._kernel_id_obj(kernel_sets)
        ] if isinstance(
            kernel_sets,
            (int, str, KernelSetDetails)
        ) else list(map(self._kernel_id_obj, kernel_sets))

    @staticmethod
    def _kernel_id_obj(kernel_set):
        # Payload kernel set object
        return {"type": "KERNEL_SET", "id": API.kernel_set_id(kernel_set)}

    @SetterProperty
    def kernel_paths(self, paths):
        '''Add path for individual kernel paths.

        Parameters
        ----------
        kernel_paths: str, [str]
            Kernel path(s) to be used for the calculation::

               [{'type': 'KERNEL', 'path': 'pds/wgc/kernels/lsk/naif0012.tls'}, ...]

        '''
        self.__kernels += [
            self._kernel_path_obj(paths)
        ] if isinstance(paths, str) else list(map(self._kernel_path_obj, paths))

    @staticmethod
    def _kernel_path_obj(server_path):
        # Payloaf individual kernel path object
        return {"type": "KERNEL", "path": server_path}

    @SetterProperty
    def times(self, times):
        '''Calculation input times.

        Parameters
        ----------
        times: str or [str]
            String or array of strings representing the time points
            that should be used in the calculation.

        Raises
        ------
        CalculationConflictAttr
            Either this parameter or the py:attr:`intervals` parameter must be supplied.

        '''
        self.__times = [times] if isinstance(times, str) else times

        if 'intervals' in self.params.keys():
            raise CalculationConflictAttr('times', 'intervals')

    @SetterProperty
    def intervals(self, intervals):
        '''Calculation input intervals.

        Parameters
        ----------
        intervals: [str, str] or {'startTime': str, 'endTime': str} or [interval, ...]
            An array of objects with startTime and endTime parameters,
            representing the time intervals used for the calculation.

            Warning
            -------
            Either this parameter or the :py:attr:`times` parameter must be supplied.

        Raises
        ------
        CalculationInvalidAttr
            If :py:attr:intervals` input format is invalid.
            For example, if :py:attr:intervals` is provided an dict,
            ``startTime`` and ``endTime`` must be present.
        CalculationUndefinedAttr
            If this parameter is used, :py:attr:`time_step` must also be supplied.

        '''
        if isinstance(intervals, dict):
            self.__intervals = [self._interval(intervals)]
        elif isinstance(intervals, list) and len(intervals) > 2:
            self.__intervals = list(map(self._interval, intervals))
        elif isinstance(intervals, list) and len(intervals) == 2:
            if isinstance(intervals[0], str) and isinstance(intervals[1], str):
                self.__intervals = [self._interval(intervals)]
            elif isinstance(intervals[0], (dict, list)) and \
                    isinstance(intervals[1], (dict, list)):
                self.__intervals = [
                    self._interval(intervals[0]), self._interval(intervals[1])]
            else:
                raise CalculationInvalidAttr('intervals', intervals, INTERVALS)
        elif isinstance(intervals, list) and len(intervals) == 1:
            if isinstance(intervals[0], (dict, list)):
                self.__intervals = [self._interval(intervals[0])]
            else:
                raise CalculationInvalidAttr('intervals', intervals, INTERVALS)
        else:
            raise CalculationInvalidAttr('intervals', intervals, INTERVALS)

        if 'time_step' not in self.params.keys():
            raise CalculationUndefinedAttr('intervals', intervals, 'time_step')

    @staticmethod
    def _interval(interval):
        # Parse interval object
        if not len(interval) == 2:
            raise CalculationInvalidAttr('interval', interval, INTERVALS[:2])

        if isinstance(interval, dict):
            if 'startTime' in interval.keys() and 'endTime' in interval.keys():
                return interval

            raise CalculationInvalidAttr('interval', interval, INTERVALS[:2])

        return {'startTime': str(interval[0]), 'endTime': str(interval[1])}

    @SetterProperty
    def time_step(self, val):
        '''Time step for intervals.

        Parameters
        ----------
        time_step: int
            Number of steps parameter used for time series or
            geometry finder calculations.

        Raises
        ------
        CalculationConflictAttr
            If :py:attr:`times` attribute is supplied.
        CalculationUndefinedAttr
            If :py:attr:`time_step_units` is not supplied.

        '''
        self.__timeStep = int(val)

        if 'times' in self.params.keys():
            raise CalculationConflictAttr('time_step', 'times')

        if 'time_step_units' not in self.params.keys():
            raise CalculationUndefinedAttr('time_step', val, 'time_step_units')

    @SetterProperty
    def time_step_units(self, val):
        '''Time step units.

        Parameters
        ----------
        time_step_units: str
            One of the following:

            - SECONDS
            - MINUTES
            - HOURS
            - DAYS
            - EQUAL_INTERVALS

        Raises
        -------
        CalculationInvalidAttr
            If the value provided is invalid.
        CalculationConflictAttr
            If :py:attr:`times` attribute is supplied.
        CalculationUndefinedAttr
            If :py:attr:`time_step` is not supplied.

        '''
        if val in TIME_STEP_UNITS:
            self.__timeStepUnits = val
        else:
            raise CalculationInvalidAttr('time_step_units', val, TIME_STEP_UNITS)

        if 'times' in self.params.keys():
            raise CalculationConflictAttr('time_step', 'times')
        if 'time_step' not in self.params.keys():
            raise CalculationUndefinedAttr('time_step_units', val, 'time_step')

    @SetterProperty
    def time_system(self, val):
        '''Time System.

        Parameters
        ----------
        time_system: str
            One of the following:

            - UTC
            - TDB
            - TDT
            - SPACECRAFT_CLOCK

        Raises
        -------
        CalculationInvalidAttr
            If the value provided is invalid.
        CalculationUndefinedAttr
            If ``SPACECRAFT_CLOCK`` is selected, but
            :py:attr:`sclk_id` attribute is not provided.

        '''
        if val in TIME_SYSTEM:
            self.__timeSystem = val
        else:
            raise CalculationInvalidAttr('time_system', val, TIME_SYSTEM)

        if val == 'SPACECRAFT_CLOCK' and 'sclk_id' not in self.params.keys():
            raise CalculationUndefinedAttr('time_system', 'SPACECRAFT_CLOCK', 'sclk_id')

    @SetterProperty
    def time_format(self, val):
        '''Time format input.

        Parameters
        ----------
        time_format: str
            One of the following:

            - CALENDAR
            - JULIAN
            - SECONDS_PAST_J2000
            - SPACECRAFT_CLOCK_TICKS
            - SPACECRAFT_CLOCK_STRING

        Raises
        -------
        CalculationInvalidAttr
            If the value provided is invalid.
        CalculationRequiredAttr
            If :py:attr:`time_system` is not provided.
        CalculationIncompatibleAttr
            If ``CALENDAR``, ``JULIAN`` or ``SECONDS_PAST_J2000`` is selected but
            :py:attr:`time_system` attribute is not in ``UTC``, ``TDB`` or ``TDT``,
            or ``SPACECRAFT_CLOCK_STRING`` or ``SPACECRAFT_CLOCK_TICKS`` is selected but
            :py:attr:`time_system` attribute is not ``SPACECRAFT_CLOCK``.

        '''
        if val in TIME_FORMAT:
            self.__timeFormat = val
        else:
            raise CalculationInvalidAttr('time_format', val, TIME_FORMAT)

        self._required(['time_system'], self.params)

        if val in ['CALENDAR', 'JULIAN', 'SECONDS_PAST_J2000'] and \
                self.params['time_system'] not in ['UTC', 'TDB', 'TDT']:
            raise CalculationIncompatibleAttr(
                'time_format', val, 'time_system', self.params['time_system'],
                ['UTC', 'TDB', 'TDT'])

        if val in ['SPACECRAFT_CLOCK_STRING', 'SPACECRAFT_CLOCK_TICKS'] and \
                self.params['time_system'] != 'SPACECRAFT_CLOCK':
            raise CalculationIncompatibleAttr(
                'time_format', val, 'time_system',
                self.params['time_system'], ['SPACECRAFT_CLOCK'])

    @SetterProperty
    def sclk_id(self, val):
        '''Spacecraft clock kernel id.

        Parameters
        ----------
        sclk_id: int
            Spacecraft clock kernel id.

        Raises
        ------
        CalculationRequiredAttr
            If :py:attr:`time_system` is not provided.
        CalculationIncompatibleAttr
            If :py:attr:`time_system` is not ``SPACECRAFT_CLOCK``.

        '''
        self.__sclkId = int(val)

        self._required(['time_system'], self.params)

        if self.params['time_system'] != 'SPACECRAFT_CLOCK':
            raise CalculationIncompatibleAttr(
                'sclk_id', val, 'time_system',
                self.params['time_system'], ['SPACECRAFT_CLOCK'])

    @SetterProperty
    def output_time_system(self, val):
        '''The time system for results output times.

        Parameters
        ----------
        output_time_system: str
            One of the following:

            - UTC
            - TDB
            - TDT
            - SPACECRAFT_CLOCK

        Raises
        -------
        CalculationInvalidAttr
            If the value provided is invalid.
        CalculationUndefinedAttr
            If ``SPACECRAFT_CLOCK`` is selected, but
            :py:attr:`output_sclk_id` attribute is not provided.

        '''
        if val in TIME_SYSTEM:
            self.__outputTimeSystem = val
        else:
            raise CalculationInvalidAttr('output_time_system', val, TIME_SYSTEM)

        if val == 'SPACECRAFT_CLOCK' and 'output_sclk_id' not in self.params.keys():
            raise CalculationUndefinedAttr(
                'output_time_system', 'SPACECRAFT_CLOCK', 'output_sclk_id')

    @SetterProperty
    def output_time_format(self, val):
        '''The time format for the result output times.

        Parameters
        ----------
        output_time_format: str
            One of the following:

            - CALENDAR
            - CALENDAR_YMD
            - CALENDAR_DOY
            - JULIAN
            - SECONDS_PAST_J2000
            - SPACECRAFT_CLOCK_STRING
            - SPACECRAFT_CLOCK_TICKS
            - CUSTOM

            Warning
            -------
            If ``CUSTOM`` is selected, then :py:attr:`output_time_custom_format`
            must also be provided.

        Raises
        -------
        CalculationInvalidAttr
            If the value provided is invalid.
        CalculationRequiredAttr
            If :py:attr:`output_time_system` is not provided.
        CalculationIncompatibleAttr
            If ``CALENDAR_YMD``, ``CALENDAR_DOY``, ``JULIAN``, ``SECONDS_PAST_J2000``
            or ``CUSTOM`` is selected but :py:attr:`outputTimeSystem` is not in
            ``TDB``, ``TDT`` or ``UTC``, or ``SPACECRAFT_CLOCK_STRING`` or
            ``SPACECRAFT_CLOCK_TICKS`` is selected but :py:attr:`output_time_system`
            is not ``SPACECRAFT_CLOCK``.

        '''
        if val in OUTPUT_TIME_FORMAT:
            self.__outputTimeFormat = val
        else:
            raise CalculationInvalidAttr('output_time_format', val, OUTPUT_TIME_FORMAT)

        self._required(['output_time_system'], self.params)

        if val in ['CALENDAR', 'CALENDAR_YMD', 'CALENDAR_DOY',
                   'JULIAN', 'SECONDS_PAST_J2000', 'CUSTOM'] and \
                self.params['output_time_system'] not in ['UTC', 'TDB', 'TDT']:
            raise CalculationIncompatibleAttr(
                'output_time_format', val,
                'output_time_system', self.params['output_time_system'],
                ['UTC', 'TDB', 'TDT'])

        if val in ['SPACECRAFT_CLOCK_STRING', 'SPACECRAFT_CLOCK_TICKS'] and \
                self.params['output_time_system'] != 'SPACECRAFT_CLOCK':
            raise CalculationIncompatibleAttr(
                'output_time_format', val,
                'output_time_system', self.params['output_time_system'],
                ['SPACECRAFT_CLOCK'])

    @SetterProperty
    def output_time_custom_format(self, val):
        '''A SPICE ``timout()`` format string.

        Parameters
        ----------
        output_time_custom_format: str
            A SPICE ``timout()`` format string.

        Raises
        ------
        CalculationRequiredAttr
            If :py:attr:`output_time_format` is not provided.
        CalculationIncompatibleAttr
            If :py:attr:`output_time_format` is not ``CUSTOM``.

        '''
        self.__outputTimeCustomFormat = val

        self._required(['output_time_format'], self.params)

        if self.params['output_time_format'] != 'CUSTOM':
            raise CalculationIncompatibleAttr(
                'output_time_custom_format', val, 'output_time_format',
                self.params['output_time_format'], ['CUSTOM'])

    @SetterProperty
    def output_sclk_id(self, val):
        '''The output spacecraft clock kernel id.

        Parameters
        ----------
        output_sclk_id: int
            Spacecraft clock kernel id.

        Raises
        ------
        CalculationRequiredAttr
            If :py:attr:`output_time_system` is not provided.
        CalculationIncompatibleAttr
            If :py:attr:`output_time_system` is not ``SPACECRAFT_CLOCK``.

        '''
        self.__outputSclkId = int(val)

        self._required(['output_time_system'], self.params)

        if self.params['output_time_system'] != 'SPACECRAFT_CLOCK':
            raise CalculationIncompatibleAttr(
                'output_sclk_id', val, 'output_time_system',
                self.params['output_time_system'], ['SPACECRAFT_CLOCK'])

    @SetterProperty
    def target(self, val):
        '''Target body.

        Parameters
        ----------
        target: str or int
            The target body ``name`` or ``id`` from :py:func:`API.bodies`.

        '''
        self.__target = val if isinstance(val, int) else val.upper()

    @SetterProperty
    def target_frame(self, val):
        '''The target body-fixed reference frame name.

        Parameters
        ----------
        target_frame: str
            Reference frame ``name``.

        '''
        self.__targetFrame = val

    @SetterProperty
    def target_1(self, val):
        '''The target body the first body.

        Parameters
        ----------
        target_1: str
            Target body ``name`` or ``id``.

        '''
        self.__target1 = val if isinstance(val, int) else val.upper()

    @SetterProperty
    def target_2(self, val):
        '''The target body the second body.

        Parameters
        ----------
        target_2: str
            Target body ``name`` or ``id``.

        '''
        self.__target2 = val if isinstance(val, int) else val.upper()

    @SetterProperty
    def shape_1(self, val):
        '''The shape to use for the first body.

        Parameters
        ----------
        shape_1: str
            One of:

            - POINT
            - SPHERE

        Raises
        -------
        CalculationInvalidAttr
            If the value provided is invalid.

        '''
        if val in SHAPE:
            self.__shape1 = val
        else:
            raise CalculationInvalidAttr('shape_1', val, SHAPE)

    @SetterProperty
    def shape_2(self, val):
        '''The shape to use for the second body.

        Parameters
        ----------
        shape_2: str
            One of:

            - POINT
            - SPHERE

        Raises
        -------
        CalculationInvalidAttr
            If the value provided is invalid.

        '''
        if val in SHAPE:
            self.__shape2 = val
        else:
            raise CalculationInvalidAttr('shape_2', val, SHAPE)

    @SetterProperty
    def observer(self, val):
        '''The observing body.

        Parameters
        ----------
        observer: str or int
            The oberving body ``name`` or ``id`` from :py:func:`API.bodies`.

        '''
        self.__observer = val if isinstance(val, int) else val.upper()

    @SetterProperty
    def reference_frame(self, val):
        '''The reference frame.

        Parameters
        ----------
        reference_frame: str or int
            The reference frame ``name`` or ``id`` from :py:func:`API.frames`.

        '''
        self.__referenceFrame = val if isinstance(val, int) else val.upper()

    @SetterProperty
    def frame_1(self, val):
        '''The first reference frame.

        Parameters
        ----------
        frame_1: str or int
            The reference frame ``name`` or ``id`` from :py:func:`API.frames`.

        '''
        self.__frame1 = val if isinstance(val, int) else val.upper()

    @SetterProperty
    def frame_2(self, val):
        '''The second reference frame.

        Parameters
        ----------
        frame_2: str or int
            The reference frame ``name`` or ``id`` from :py:func:`API.frames`.

        '''
        self.__frame2 = val if isinstance(val, int) else val.upper()

    @SetterProperty
    def orbiting_body(self, val):
        '''The SPICE orbiting body.

        Parameters
        ----------
        orbiting_body: str or int
            SPICE body ``name`` or ``id`` for the orbiting body.

        '''
        self.__orbitingBody = val if isinstance(val, int) else val.upper()

    @SetterProperty
    def center_body(self, val):
        '''
        The SPICE body center of motion.

        Parameters
        ----------
        center_body: str or int
            SPICE body ``name`` or ``id`` for the body that is the center of motion.

        '''
        self.__centerBody = val if isinstance(val, int) else val.upper()

    @SetterProperty
    def aberration_correction(self, val):
        '''SPICE aberration correction.

        Parameters
        ----------
        aberration_correction: str
            The SPICE aberration correction string. One of:

            - NONE
            - LT
            - LT+S
            - CN
            - CN+S
            - XLT
            - XLT+S
            - XCN
            - XCN+S

        Raises
        ------
        CalculationInvalidAttr
            If the value provided is invalid.

        '''
        if val in ABERRATION_CORRECTION:
            self.__aberrationCorrection = val
        else:
            raise CalculationInvalidAttr(
                'aberration_correction', val, ABERRATION_CORRECTION)

    @SetterProperty
    def state_representation(self, val):
        '''State representation.

        Parameters
        ----------
        state_representation: str
            One of:

            - RECTANGULAR
            - RA_DEC
            - LATITUDINAL (planetocentric)
            - PLANETODETIC
            - PLANETOGRAPHIC
            - CYLINDRICAL
            - SPHERICAL

        Raises
        ------
        CalculationInvalidAttr
            If the value provided is invalid.

        '''
        if val in STATE_REPRESENTATION:
            self.__stateRepresentation = val
        else:
            raise CalculationInvalidAttr(
                'state_representation', val, STATE_REPRESENTATION)

    @SetterProperty
    def time_location(self, val):
        '''The frame for the input times.

        Parameters
        ----------
        time_location: str
            One of:

            - FRAME1
            - FRAME2

        Raises
        ------
        CalculationInvalidAttr
            If the value provided is invalid.

        Warning
        -------
        NAIF API docs::

            `Only needed if aberrationCorrection is not NONE.`

        Required even when :py:attr:`aberration_correction` is ``NONE``.

        '''
        if val in TIME_LOCATION:
            self.__timeLocation = val
        else:
            raise CalculationInvalidAttr('time_location', val, TIME_LOCATION)

    @SetterProperty
    def orientation_representation(self, val):
        '''The representation of the result transformation.

        Parameters
        ----------
        orientation_representation: str
            Orientation result transformation. One of:

            - EULER_ANGLES
            - ANGLE_AND_AXIS
            - SPICE_QUATERNION
            - OTHER_QUATERNION
            - MATRIX_ROW_BY_ROW
            - MATRIX_FLAGGED
            - MATRIX_ALL_ONE_ROW

        Raises
        ------
        CalculationInvalidAttr
            If the value provided is invalid.

        '''
        if val in ORIENTATION_REPRESENTATION:
            self.__orientationRepresentation = val
        else:
            raise CalculationInvalidAttr(
                'orientation_representation', val, ORIENTATION_REPRESENTATION)

    @SetterProperty
    def axis_1(self, val):
        '''The first axis for Euler angle rotation.

        Parameters
        ----------
        axis_1: str
            Axis name. See: :py:func:`axis`.

        '''
        self.__axis1 = self.axis('axis_1', val)

    @SetterProperty
    def axis_2(self, val):
        '''The second axis for Euler angle rotation.

        Parameters
        ----------
        axis_3: str
            Axis name. See: :py:func:`axis`.

        '''
        self.__axis2 = self.axis('axis_2', val)

    @SetterProperty
    def axis_3(self, val):
        '''The third axis for Euler angle rotation.

        Parameters
        ----------
        axis_3: str
            Axis name. See: :py:func:`axis`.

        '''
        self.__axis3 = self.axis('axis_3', val)

    def axis(self, name, val):
        '''Axis for Euler angle rotation.

        Parameters
        ----------
        name: str
            Axis name. One of:

            - X
            - Y
            - Z

        val: float
            Value on the axis.

        Return
        ------
        float
            Value on the axis.

        Raises
        ------
        CalculationInvalidAttr
            If the ``name`` provided is invalid.
        CalculationUndefinedAttr
            If :py:attr:`orientation_representation` is not supplied.
        CalculationIncompatibleAttr
            If :py:attr:`orientation_representation` is not ``EULER_ANGLES``.

        '''
        if 'orientation_representation' not in self.params.keys():
            raise CalculationUndefinedAttr(name, val, 'orientation_representation')

        if self.params['orientation_representation'] != 'EULER_ANGLES':
            raise CalculationIncompatibleAttr(
                name, val, 'orientation_representation',
                self.params['orientation_representation'], ['EULER_ANGLES'])

        if val in AXIS:
            return val

        raise CalculationInvalidAttr(name, val, AXIS)

    @SetterProperty
    def angular_units(self, val):
        '''The angular units.

        Parameters
        ----------
        angular_units: str
            The angular units used for the angle of rotation. One of:

            - deg
            - rad

        Raises
        ------
        CalculationInvalidAttr
            If the value provided is invalid.
        CalculationUndefinedAttr
            If :py:attr:`orientation_representation` is not supplied.
        CalculationIncompatibleAttr
            If :py:attr:`orientation_representation` is
            not `EULER_ANGLES` or `ANGLE_AND_AXIS`.

        '''
        if val in ANGULAR_UNITS:
            self.__angularUnits = val
        else:
            raise CalculationInvalidAttr('angular_units', val, ANGULAR_UNITS)

        if 'orientation_representation' not in self.params.keys():
            raise CalculationUndefinedAttr(
                'angular_units', val, 'orientation_representation')

        if not self.params['orientation_representation'] in ['EULER_ANGLES',
                                                             'ANGLE_AND_AXIS']:
            raise CalculationIncompatibleAttr(
                'angular_units', val, 'orientation_representation',
                self.params['orientation_representation'],
                ['EULER_ANGLES', 'ANGLE_AND_AXIS'])

    @SetterProperty
    def angular_velocity_representation(self, val):
        '''Angular velocity representation.

        Parameters
        ----------
        angular_velocity_representation: str
            The representation of angular velocity in the output. One of:

            - NOT_INCLUDED
            - VECTOR_IN_FRAME1
            - VECTOR_IN_FRAME2
            - EULER_ANGLE_DERIVATIVES
            - MATRIX

        Raises
        ------
        CalculationInvalidAttr
            If the value provided is invalid.

        '''
        if val in ANGULAR_VELOCITY_REPRESENTATION:
            self.__angularVelocityRepresentation = val
        else:
            raise CalculationInvalidAttr(
                'angular_velocity_representation', val, ANGULAR_VELOCITY_REPRESENTATION)

    @SetterProperty
    def angular_velocity_units(self, val):
        '''The units for the angular velocity.

        Parameters
        ----------
        angular_velocity_units: str
            One of:

            - deg/s
            - rad/s
            - RPM
            - Unitary *(ie, Unit vector)*

        Raises
        ------
        CalculationInvalidAttr
            If the value provided is invalid.
        CalculationUndefinedAttr
            If :py:attr:`angular_velocity_representation` is not supplied.
        CalculationIncompatibleAttr
            If :py:attr:`angular_velocity_representation` is not ``VECTOR_IN_FRAME1``,
            ``VECTOR_IN_FRAME2`` or ``EULER_ANGLE_DERIVATIVES``
        CalculationIncompatibleAttr
            If ``Unitary`` selected but :py:attr:`angular_velocity_representation` is not
            ``VECTOR_IN_FRAME1`` or ``VECTOR_IN_FRAME2``.

        '''
        if val in ANGULAR_VELOCITY_UNITS:
            self.__angularVelocityUnits = val
        else:
            raise CalculationInvalidAttr(
                'angular_velocity_units', val, ANGULAR_VELOCITY_UNITS)

        if 'angular_velocity_representation' not in self.params.keys():
            raise CalculationUndefinedAttr(
                'angular_velocity_units', val, 'angular_velocity_representation')

        choices = ['VECTOR_IN_FRAME1', 'VECTOR_IN_FRAME2', 'EULER_ANGLE_DERIVATIVES']
        if not self.params['angular_velocity_representation'] in choices:
            raise CalculationIncompatibleAttr(
                'angular_velocity_units', val, 'angular_velocity_representation',
                self.params['angular_velocity_representation'], choices)

        if val == 'Unitary' and not self.params['angular_velocity_representation'] in \
                ['VECTOR_IN_FRAME1', 'VECTOR_IN_FRAME2']:
            raise CalculationIncompatibleAttr(
                'angular_velocity_units', val, 'angular_velocity_representation',
                self.params['angular_velocity_representation'],
                ['VECTOR_IN_FRAME1', 'VECTOR_IN_FRAME2'])

    @SetterProperty
    def coordinate_representation(self, val):
        '''Coordinate Representation.

        Parameters
        ----------
        coordinate_representation: str
            One of:

            - LATITUDINAL *(planetocentric)*
            - PLANETODETIC
            - PLANETOGRAPHIC

        Raises
        ------
        CalculationInvalidAttr
            If the value provided is invalid.

        '''
        if val in COORDINATE_REPRESENTATION:
            self.__coordinateRepresentation = val
        else:
            raise CalculationInvalidAttr(
                'coordinate_representation', val, COORDINATE_REPRESENTATION)

    @SetterProperty
    def latitude(self, val):
        '''Latitude of the surface point.

        Parameters
        ----------
        latitude: float
            Latitude angle (in degrees).

        Raises
        ------
        CalculationInvalidValue
            If ``latitude`` not in [-90, +90] range.

        '''
        if -90 <= val <= 90:
            self.__latitude = val
        else:
            raise CalculationInvalidValue('latitude', val, -90, 90)

    @SetterProperty
    def longitude(self, val):
        '''Longitude of the surface point.

        Parameters
        ----------
        longitude: float
            Longitude angle (in degrees).

        Raises
        ------
        CalculationInvalidValue
            If ``longitude`` not in [-180, +180] range.

        '''
        if -180 <= val <= 180:
            self.__longitude = val
        else:
            raise CalculationInvalidValue('longitude', val, -180, 180)

    @SetterProperty
    def sub_point_type(self, val):
        '''Sub-observer point.

        Parameters
        ----------
        sub_point_type: str
            The method of finding the sub-observer point, as in
            the SPICE ``subpnt()`` API call. One of:

            - Near point: ellipsoid
            - Intercept: ellipsoid
            - NADIR/DSK/UNPRIORITIZED
            - INTERCEPT/DSK/UNPRIORITIZED

        Raises
        ------
        CalculationInvalidAttr
            If the value provided is invalid.

        '''
        if val in SUB_POINT_TYPE:
            self.__subPointType = val
        else:
            raise CalculationInvalidAttr('sub_point_type', val, SUB_POINT_TYPE)

    @SetterProperty
    def direction_vector_type(self, val):
        '''Type of ray's direction vector.

        Parameters
        ----------
        direction_vector_type: str
            Type of vector to be used as the ray direction. One of:

            - INSTRUMENT_BORESIGHT *(the instrument boresight vector)*
            - INSTRUMENT_FOV_BOUNDARY_VECTORS *(the instrument field-of-view
              boundary vectors)*
            - REFERENCE_FRAME_AXIS *(an axis of the specified reference frame)*
            - VECTOR_IN_INSTRUMENT_FOV *(a vector in the reference frame of the
              specified instrument)*
            - VECTOR_IN_REFERENCE_FRAME *(a vector in the specified reference frame)*

        Raises
        ------
        CalculationInvalidAttr
            If the value provided is invalid.
        CalculationRequiredAttr
            If this parameter is ``INSTRUMENT_BORESIGHT``,
            ``INSTRUMENT_FOV_BOUNDARY_VECTORS`` or ``VECTOR_IN_INSTRUMENT_FOV``
            but :py:attr:`direction_instrument` is not provided.
        CalculationRequiredAttr
            If this parameter is ``REFERENCE_FRAME_AXIS`` or ``VECTOR_IN_REFERENCE_FRAME``
            but :py:attr:`direction_frame` is not provided.
        CalculationRequiredAttr
            If this parameter is ``REFERENCE_FRAME_AXIS`` but
            :py:attr:`direction_frame_axis` is not provided.
        CalculationUndefinedAttr
            If this parameter is ``VECTOR_IN_INSTRUMENT_FOV``
            or ``VECTOR_IN_REFERENCE_FRAME`` but neither :py:attr:`direction_vector_x`,
            :py:attr:`direction_vector_y` and :py:attr:`direction_vector_z`
            nor :py:attr:`direction_vector_ra` and :py:attr:`direction_vector_dec`
            are provided.

        '''
        if val in DIRECTION_VECTOR_TYPE:
            self.__directionVectorType = val
        else:
            raise CalculationInvalidAttr(
                'direction_vector_type', val, DIRECTION_VECTOR_TYPE)

        if val in ['INSTRUMENT_BORESIGHT', 'INSTRUMENT_FOV_BOUNDARY_VECTORS',
                   'VECTOR_IN_INSTRUMENT_FOV']:
            self._required(['direction_instrument'], self.params)

        elif val in ['REFERENCE_FRAME_AXIS', 'VECTOR_IN_REFERENCE_FRAME']:
            self._required(['direction_frame'], self.params)
            if val == 'REFERENCE_FRAME_AXIS':
                self._required(['direction_frame_axis'], self.params)

        keys = self.params.keys()
        if val in ['VECTOR_IN_INSTRUMENT_FOV', 'VECTOR_IN_REFERENCE_FRAME']:
            if not(
                    'direction_vector_x' in keys and  # noqa: W504
                    'direction_vector_y' in keys and  # noqa: W504
                    'direction_vector_z' in keys
            ) and not(
                'direction_vector_ra' in keys and  # noqa: W504
                'direction_vector_dec' in keys
            ):
                raise CalculationUndefinedAttr(
                    'direction_vector_type', val,
                    "direction_vector_x/y/z' or 'direction_vector_ra/dec")

    @SetterProperty
    def direction_instrument(self, val):
        '''Direction imnstrument.

        Parameters
        ----------
        direction_instrument: str or int
            The instrument ``name`` or ``id``.

        Raises
        ------
        CalculationUndefinedAttr
            If :py:attr:`direction_vector_type` is not provided.
        CalculationIncompatibleAttr
            If :py:attr:`direction_vector_type` not in ``INSTRUMENT_BORESIGHT``,
            ``INSTRUMENT_FOV_BOUNDARY_VECTORS`` or ``VECTOR_IN_INSTRUMENT_FOV``.

        '''
        if 'direction_vector_type' not in self.params.keys():
            raise CalculationUndefinedAttr(
                'direction_instrument', val, 'direction_vector_type')

        choices = ['INSTRUMENT_BORESIGHT', 'INSTRUMENT_FOV_BOUNDARY_VECTORS',
                   'VECTOR_IN_INSTRUMENT_FOV']
        if not self.params['direction_vector_type'] in choices:
            raise CalculationIncompatibleAttr(
                'direction_instrument', val, 'direction_vector_type',
                self.params['direction_vector_type'], choices)

        self.__directionInstrument = val if isinstance(val, int) else val.upper()

    @SetterProperty
    def direction_frame(self, val):
        '''Direction vector reference frame.

        Parameters
        ----------
        direction_frame: str
            The vector's reference frame ``name``.

        Raises
        ------
        CalculationUndefinedAttr
            If :py:attr:`direction_vector_type` is not provided.
        CalculationIncompatibleAttr
            If :py:attr:`direction_vector_type` is not in ``REFERENCE_FRAME_AXIS``
            or ``VECTOR_IN_REFERENCE_FRAME``.

        '''
        if 'direction_vector_type' not in self.params.keys():
            raise CalculationUndefinedAttr(
                'direction_frame', val, 'direction_vector_type')

        choices = ['REFERENCE_FRAME_AXIS', 'VECTOR_IN_REFERENCE_FRAME']
        if not self.params['direction_vector_type'] in choices:
            raise CalculationIncompatibleAttr(
                'direction_frame', val, 'direction_vector_type',
                self.params['direction_vector_type'], choices)

        self.__directionFrame = val

    @SetterProperty
    def direction_frame_axis(self, val):
        '''The vector's reference frame axis name.

        Parameters
        ----------
        direction_frame: str
            The vector's reference frame  axis. One of:

            - X
            - Y
            - Z

        Raises
        ------
        CalculationUndefinedAttr
            If :py:attr:`direction_vector_type` is not provided.
        CalculationIncompatibleAttr
            If :py:attr:`direction_vector_type` is not ``REFERENCE_FRAME_AXIS``.
        CalculationInvalidAttr
            If the value provided is invalid.

        '''
        if 'direction_vector_type' not in self.params.keys():
            raise CalculationUndefinedAttr(
                'direction_frame_axis', val, 'direction_vector_type')

        if self.params['direction_vector_type'] != 'REFERENCE_FRAME_AXIS':
            raise CalculationIncompatibleAttr(
                'direction_frame_axis', val, 'direction_vector_type',
                self.params['direction_vector_type'], ['REFERENCE_FRAME_AXIS'])

        if val in AXIS:
            self.__directionFrameAxis = val
        else:
            raise CalculationInvalidAttr('direction_frame_axis', val, AXIS)

    def direction_vector(self, axis, val):
        '''Direction vector coordinate.

        Parameters
        ----------
        axis: str
            Axis name.
        val: float
            Value on the axis.

        Raises
        ------
        CalculationUndefinedAttr
            If :py:attr:`direction_vector_type` is not provided.
        CalculationIncompatibleAttr
            If :py:attr:`direction_vector_type` is not in ``VECTOR_IN_INSTRUMENT_FOV``
            or ``VECTOR_IN_REFERENCE_FRAME``.

        '''
        if 'direction_vector_type' not in self.params.keys():
            raise CalculationUndefinedAttr(
                'direction_vector_' + axis, val, 'direction_vector_type')

        choices = ['VECTOR_IN_INSTRUMENT_FOV', 'VECTOR_IN_REFERENCE_FRAME']
        if not self.params['direction_vector_type'] in choices:
            raise CalculationIncompatibleAttr(
                'direction_vector_' + axis, val, 'direction_vector_type',
                self.params['direction_vector_type'], choices)
        return val

    @SetterProperty
    def direction_vector_x(self, val):
        '''The X ray's direction vector coordinate.

        Parameters
        ----------
        direction_vector_x: float
            Direction x-coordinate. See :py:func:`direction_vector`.

        '''
        self.__directionVectorX = self.direction_vector('x', val)

    @SetterProperty
    def direction_vector_y(self, val):
        '''The Y ray's direction vector coordinate.

        Parameters
        ----------
        direction_vector_y: float
            Direction y-coordinate. See :py:func:`direction_vector`.

        '''
        self.__directionVectorY = self.direction_vector('y', val)

    @SetterProperty
    def direction_vector_z(self, val):
        '''The Z ray's direction vector coordinate.

        Parameters
        ----------
        direction_vector_z: float
            Direction z-coordinate. See :py:func:`direction_vector`.

        '''
        self.__directionVectorZ = self.direction_vector('z', val)

    @SetterProperty
    def direction_vector_ra(self, val):
        '''The right-ascenssion ray's direction vector coordinate.

        Parameters
        ----------
        direction_vector_ra: float
            Direction RA-coordinate. See :py:func:`direction_vector`.

        '''
        self.__directionVectorRA = self.direction_vector('ra', val)

    @SetterProperty
    def direction_vector_dec(self, val):
        '''The declination ray's direction vector coordinate.

        Parameters
        ----------
        direction_vector_dec: float
            Direction DEC-coordinate. See :py:func:`direction_vector`.

        '''
        self.__directionVectorDec = self.direction_vector('dec', val)



# -*- coding: utf-8 -*-
'''Webgeocalc calculations types.'''

from .calculation import Calculation
from .errors import CalculationInvalidAttr
from .vars import ABERRATION_CORRECTION


class StateVector(Calculation):
    '''State vector calculation.

    Calculates the position of one body relative to another,
    calculated in a desired reference frame.

    Parameters
    ----------
    aberration_correction: str, optional
        See: :py:attr:`.aberration_correction`
    state_representation: str, optional
        See: :py:attr:`.state_representation`

    Other Parameters
    ----------------
    kernels: str, int, [str or/and int]
        See: :py:attr:`.kernels`
    kernel_paths: str, [str]
        See: :py:attr:`.kernel_paths`
    times: str or [str]
        See: :py:attr:`.times`
    intervals: [str, str] or {'startTime': str, 'endTime': str} or [interval, ...]
        See: :py:attr:`.intervals`
    time_step: int
        See: :py:attr:`.time_step`
    time_step_units: str
        See: :py:attr:`.time_step_units`
    time_system: str
        See: :py:attr:`.time_system`
    time_format: str
        See: :py:attr:`.time_format`
    target: str or int
        See: :py:attr:`.target`
    observer: str or int
        See: :py:attr:`.observer`
    reference_frame: str or int
        See: :py:attr:`.reference_frame`

    Raises
    ------
    CalculationRequiredAttr
        If :py:attr:`.target`, :py:attr:`.observer` and
        :py:attr:`.reference_frame` are not provided.

    '''

    def __init__(self, aberration_correction='CN',
                 state_representation='RECTANGULAR', **kwargs):

        self._required(['target', 'observer', 'reference_frame'], kwargs)

        kwargs['calculation_type'] = 'STATE_VECTOR'
        kwargs['aberration_correction'] = aberration_correction
        kwargs['state_representation'] = state_representation

        super().__init__(**kwargs)

class AngularSeparation(Calculation):
    '''Angular separation calculation.

    Calculates the angular separation of two bodies as seen by an observer body.

    Parameters
    ----------
    shape_1: str, optional
        See: :py:attr:`.shape_1`
    shape_2: str, optional
        See: :py:attr:`.shape_2`
    aberration_correction: str, optional
        See: :py:attr:`.aberration_correction`

    Other Parameters
    ----------------
    kernels: str, int, [str or/and int]
        See: :py:attr:`.kernels`
    kernel_paths: str, [str]
        See: :py:attr:`.kernel_paths`
    times: str or [str]
        See: :py:attr:`.times`
    intervals: [str, str] or {'startTime': str, 'endTime': str} or [interval, ...]
        See: :py:attr:`.intervals`
    time_step: int
        See: :py:attr:`.time_step`
    time_step_units: str
        See: :py:attr:`.time_step_units`
    time_system: str
        See: :py:attr:`.time_system`
    time_format: str
        See: :py:attr:`.time_format`
    target_1: str or int
        See: :py:attr:`.target_1`
    target_2: str or int
        See: :py:attr:`.target_2`
    observer: str or int
        See: :py:attr:`.observer`

    Raises
    ------
    CalculationRequiredAttr
        If :py:attr:`.target_1`, :py:attr:`.target_2` and
        :py:attr:`.observer` are not provided.

    '''

    def __init__(self, shape_1='POINT', shape_2='POINT',
                 aberration_correction='CN', **kwargs):

        self._required(['target_1', 'target_2', 'observer'], kwargs)

        kwargs['calculation_type'] = 'ANGULAR_SEPARATION'
        kwargs['shape_1'] = shape_1
        kwargs['shape_2'] = shape_2
        kwargs['aberration_correction'] = aberration_correction

        super().__init__(**kwargs)

class AngularSize(Calculation):
    '''Angular size calculation.

    Calculates the angular size of a target as seen by an observer.

    Parameters
    ----------
    aberration_correction: str, optional
        See: :py:attr:`.aberration_correction`

    Other Parameters
    ----------------
    kernels: str, int, [str or/and int]
        See: :py:attr:`.kernels`
    kernel_paths: str, [str]
        See: :py:attr:`.kernel_paths`
    times: str or [str]
        See: :py:attr:`.times`
    intervals: [str, str] or {'startTime': str, 'endTime': str} or [interval, ...]
        See: :py:attr:`.intervals`
    time_step: int
        See: :py:attr:`.time_step`
    time_step_units: str
        See: :py:attr:`.time_step_units`
    time_system: str
        See: :py:attr:`.time_system`
    time_format: str
        See: :py:attr:`.time_format`
    target: str or int
        See: :py:attr:`.target`
    observer: str or int
        See: :py:attr:`.observer`

    Raises
    ------
    CalculationRequiredAttr
        If :py:attr:`.target` and :py:attr:`.observer` are not provided.

    '''

    def __init__(self, aberration_correction='CN', **kwargs):

        self._required(['target', 'observer'], kwargs)

        kwargs['calculation_type'] = 'ANGULAR_SIZE'
        kwargs['aberration_correction'] = aberration_correction

        super().__init__(**kwargs)


class FrameTransformation(Calculation):
    '''Frame transforme calculation.

    Calculate the transformation from one reference frame (Frame 1)
    to another reference frame (Frame 2).

    Parameters
    ----------
    aberration_correction: str, optional
        See: :py:attr:`.aberration_correction`
    time_location: str, optional
        See: :py:attr:`.time_location`
    orientation_representation: str, optional
        See: :py:attr:`.orientation_representation`
    axis_1: str, optional
        See: :py:attr:`.axis_1`
    axis_2: str, optional
        See: :py:attr:`.axis_2`
    axis_3: str, optional
        See: :py:attr:`.axis_3`
    angular_units: str, optional
        See: :py:attr:`.angular_units`
    angular_velocity_representation: str, optional
        See: :py:attr:`.angular_velocity_representation`
    angular_velocity_units: str, optional
        See: :py:attr:`.angular_velocity_units`

        Warning
        -------
        Attribute :py:attr:`.aberration_correction` must be ``NONE``, `LT``, ``CN``,
        ``XLT`` or ``XCN``.

        Attributes :py:attr:`.axis_1`, :py:attr:`.axis_2` and :py:attr:`.axis_3`
        are imported only if :py:attr:`.orientation_representation` is ``EULER_ANGLES``.

        Attribute :py:attr:`.angular_units` is imported only
        if :py:attr:`.orientation_representation` is ``EULER_ANGLES``
        or ``ANGLE_AND_AXIS``.

        Attribute :py:attr:`.angular_velocity_units` is imported only if
        :py:attr:`.angular_velocity_representation` is ``VECTOR_IN_FRAME1``,
        ``VECTOR_IN_FRAME2`` or ``EULER_ANGLE_DERIVATIVES``.

    Other Parameters
    ----------------
    kernels: str, int, [str or/and int]
        See: :py:attr:`.kernels`
    kernel_paths: str, [str]
        See: :py:attr:`.kernel_paths`
    times: str or [str]
        See: :py:attr:`.times`
    intervals: [str, str] or {'startTime': str, 'endTime': str} or [interval, ...]
        See: :py:attr:`.intervals`
    time_step: int
        See: :py:attr:`.time_step`
    time_step_units: str
        See: :py:attr:`.time_step_units`
    time_system: str
        See: :py:attr:`.time_system`
    time_format: str
        See: :py:attr:`.time_format`
    frame_1: str or int
        See: :py:attr:`.frame_1`
    frame_2: str or int
        See: :py:attr:`.frame_2`

    Raises
    ------
    CalculationRequiredAttr
        If :py:attr:`.frame_1` and :py:attr:`.frame_2` are not provided.
    CalculationInvalidAttr
        If :py:attr:`.aberration_correction` is in ``LT+S``, ``CN+S``,
        ``XLT+S`` or ``XCN+S``.

    '''

    def __init__(self, aberration_correction='CN', time_location='FRAME1',
                 orientation_representation='EULER_ANGLES',
                 axis_1='X', axis_2='Y', axis_3='Z',
                 angular_units='deg',
                 angular_velocity_representation='VECTOR_IN_FRAME1',
                 angular_velocity_units='deg/s',
                 **kwargs):

        self._required(['frame_1', 'frame_2'], kwargs)

        aberration_correction_allowed = list(filter(
            lambda x: '+S' not in x, ABERRATION_CORRECTION))

        if aberration_correction not in aberration_correction_allowed:
            raise CalculationInvalidAttr(
                'aberration_correction', aberration_correction,
                aberration_correction_allowed)

        kwargs['calculation_type'] = 'FRAME_TRANSFORMATION'
        kwargs['aberration_correction'] = aberration_correction
        kwargs['time_location'] = time_location
        kwargs['orientation_representation'] = orientation_representation
        kwargs['angular_velocity_representation'] = angular_velocity_representation

        if orientation_representation == 'EULER_ANGLES':
            kwargs['axis_1'] = axis_1
            kwargs['axis_2'] = axis_2
            kwargs['axis_3'] = axis_3

        if orientation_representation in ['EULER_ANGLES', 'ANGLE_AND_AXIS']:
            kwargs['angular_units'] = angular_units

        if angular_velocity_representation in ['VECTOR_IN_FRAME1', 'VECTOR_IN_FRAME2',
                                               'EULER_ANGLE_DERIVATIVES']:
            kwargs['angular_velocity_units'] = angular_velocity_units

        super().__init__(**kwargs)


class IlluminationAngles(Calculation):
    '''Illumination angles calculation.

    Calculate the emission, phase and solar incidence angles
    at a point on a target as seen from an observer.

    Parameters
    ----------
    shape_1: str, optional
        See: :py:attr:`.shape_1`
    coordinate_representation: str, optional
        See: :py:attr:`.coordinate_representation`
    aberration_correction: str, optional
        See: :py:attr:`.aberration_correction`

    Other Parameters
    ----------------
    kernels: str, int, [str or/and int]
        See: :py:attr:`.kernels`
    kernel_paths: str, [str]
        See: :py:attr:`.kernel_paths`
    times: str or [str]
        See: :py:attr:`.times`
    intervals: [str, str] or {'startTime': str, 'endTime': str} or [interval, ...]
        See: :py:attr:`.intervals`
    time_step: int
        See: :py:attr:`.time_step`
    time_step_units: str
        See: :py:attr:`.time_step_units`
    time_system: str
        See: :py:attr:`.time_system`
    time_format: str
        See: :py:attr:`.time_format`
    target: str or int
        See: :py:attr:`.target`
    target_frame: str or int
        See: :py:attr:`.target_frame`
    observer: str or int
        See: :py:attr:`.observer`
    latitude: str or int
        See: :py:attr:`.latitude`
    longitude: str or int
        See: :py:attr:`.longitude`

    Raises
    ------
    CalculationRequiredAttr
        If :py:attr:`.target`, :py:attr:`.target_frame`, :py:attr:`.observer`,
        :py:attr:`.latitude` and :py:attr:`.longitude` are not provided.
    CalculationInvalidAttr
        If :py:attr:`.shape_1` is not ``ELLIPSOID`` or ``DSK``.

    '''

    def __init__(self, shape_1='ELLIPSOID', coordinate_representation='LATITUDINAL',
                 aberration_correction='CN', **kwargs):

        self._required(['target', 'target_frame', 'observer',
                        'latitude', 'longitude'], kwargs)

        kwargs['calculation_type'] = 'ILLUMINATION_ANGLES'
        kwargs['coordinate_representation'] = coordinate_representation
        kwargs['aberration_correction'] = aberration_correction

        if shape_1 in ['ELLIPSOID', 'DSK']:
            kwargs['shape_1'] = shape_1
        else:
            raise CalculationInvalidAttr('shape_1', shape_1, ['ELLIPSOID', 'DSK'])

        super().__init__(**kwargs)


class SubSolarPoint(Calculation):
    '''Sub-solar point calculation.

    Calculates the sub-solar point on a target as seen from an observer.

    Parameters
    ----------
    sub_point_type: str, optional
        See: :py:attr:`.sub_point_type`
    aberration_correction: str, optional
        See: :py:attr:`.aberration_correction`
    state_representation: str, optional
        See: :py:attr:`.state_representation`

        Warning
        -------
        Attribute :py:attr:`.aberration_correction` must be ``NONE``, `LT``, ``LT+S``,
        ``CN`` or ``CN+S``.

    Other Parameters
    ----------------
    kernels: str, int, [str or/and int]
        See: :py:attr:`.kernels`
    kernel_paths: str, [str]
        See: :py:attr:`.kernel_paths`
    times: str or [str]
        See: :py:attr:`.times`
    intervals: [str, str] or {'startTime': str, 'endTime': str} or [interval, ...]
        See: :py:attr:`.intervals`
    time_step: int
        See: :py:attr:`.time_step`
    time_step_units: str
        See: :py:attr:`.time_step_units`
    time_system: str
        See: :py:attr:`.time_system`
    time_format: str
        See: :py:attr:`.time_format`
    target: str or int
        See: :py:attr:`.target`
    target_frame: str or int
        See: :py:attr:`.target_frame`
    observer: str or int
        See: :py:attr:`.observer`

    Raises
    ------
    CalculationRequiredAttr
        If :py:attr:`.target`, :py:attr:`.target_frame` and :py:attr:`.observer`
        are not provided.
    CalculationInvalidAttr
        If :py:attr:`.aberration_correction` is in ``XLT``, ``XLT+S``,
        ``XCN+S`` or ``XCN+S``.

    '''

    def __init__(self, sub_point_type='Near point: ellipsoid', aberration_correction='CN',
                 state_representation='RECTANGULAR', **kwargs):

        self._required(['target', 'target_frame', 'observer'], kwargs)

        aberration_correction_allowed = list(filter(
            lambda x: 'X' not in x, ABERRATION_CORRECTION))

        if aberration_correction not in aberration_correction_allowed:
            raise CalculationInvalidAttr(
                'aberration_correction', aberration_correction,
                aberration_correction_allowed)

        kwargs['calculation_type'] = 'SUB_SOLAR_POINT'
        kwargs['sub_point_type'] = sub_point_type
        kwargs['aberration_correction'] = aberration_correction
        kwargs['state_representation'] = state_representation

        super().__init__(**kwargs)

class SubObserverPoint(Calculation):
    '''Sub-observer point calculation.

    Calculate the sub-observer point on a target as seen from an observer.

    Parameters
    ----------
    sub_point_type: str, optional
        See: :py:attr:`.sub_point_type`
    aberration_correction: str, optional
        See: :py:attr:`.aberration_correction`
    state_representation: str, optional
        See: :py:attr:`.state_representation`

    Other Parameters
    ----------------
    kernels: str, int, [str or/and int]
        See: :py:attr:`.kernels`
    kernel_paths: str, [str]
        See: :py:attr:`.kernel_paths`
    times: str or [str]
        See: :py:attr:`.times`
    intervals: [str, str] or {'startTime': str, 'endTime': str} or [interval, ...]
        See: :py:attr:`.intervals`
    time_step: int
        See: :py:attr:`.time_step`
    time_step_units: str
        See: :py:attr:`.time_step_units`
    time_system: str
        See: :py:attr:`.time_system`
    time_format: str
        See: :py:attr:`.time_format`
    target: str or int
        See: :py:attr:`.target`
    target_frame: str or int
        See: :py:attr:`.target_frame`
    observer: str or int
        See: :py:attr:`.observer`

    Raises
    ------
    CalculationRequiredAttr
        If :py:attr:`.target`, :py:attr:`.target_frame`
        and :py:attr:`.observer` are not provided.

    '''

    def __init__(self, sub_point_type='Near point: ellipsoid', aberration_correction='CN',
                 state_representation='RECTANGULAR', **kwargs):

        self._required(['target', 'target_frame', 'observer'], kwargs)

        kwargs['calculation_type'] = 'SUB_OBSERVER_POINT'
        kwargs['sub_point_type'] = sub_point_type
        kwargs['aberration_correction'] = aberration_correction
        kwargs['state_representation'] = state_representation

        super().__init__(**kwargs)


class SurfaceInterceptPoint(Calculation):
    '''Surface intercept point calculation.

    Calculate the intercept point of a vector or vectors
    on a target as seen from an observer.

    Parameters
    ----------
    shape_1: str, optional
        See: :py:attr:`.shape_1`
    direction_vector_type: str, optional
        See: :py:attr:`.direction_vector_type`
    aberration_correction: str, optional
        See: :py:attr:`.aberration_correction`
    state_representation: str, optional
        See: :py:attr:`.state_representation`

    Other Parameters
    ----------------
    kernels: str, int, [str or/and int]
        See: :py:attr:`.kernels`
    kernel_paths: str, [str]
        See: :py:attr:`.kernel_paths`
    times: str or [str]
        See: :py:attr:`.times`
    intervals: [str, str] or {'startTime': str, 'endTime': str} or [interval, ...]
        See: :py:attr:`.intervals`
    time_step: int
        See: :py:attr:`.time_step`
    time_step_units: str
        See: :py:attr:`.time_step_units`
    time_system: str
        See: :py:attr:`.time_system`
    time_format: str
        See: :py:attr:`.time_format`
    target: str or int
        See: :py:attr:`.target`
    target_frame: str or int
        See: :py:attr:`.target_frame`
    observer: str or int
        See: :py:attr:`.observer`
    direction_instrument: str or int
        See: :py:attr:`.direction_instrument`
    direction_frame: str
        See: :py:attr:`.direction_frame`
    direction_frame_axis: str
        See: :py:attr:`.direction_frame_axis`
    direction_vector_x: float
        See: :py:attr:`.direction_vector_x`
    direction_vector_y: float
        See: :py:attr:`.direction_vector_y`
    direction_vector_z: float
        See: :py:attr:`.direction_vector_z`
    direction_vector_ra: float
        See: :py:attr:`.direction_vector_ra`
    direction_vector_dec: float
        See: :py:attr:`.direction_vector_dec`

        Warnings
        --------
        Attributes :py:attr:`.direction_instrument` is needed only if
        :py:attr:`.direction_vector_type` is ``INSTRUMENT_BORESIGHT``,
        ``INSTRUMENT_FOV_BOUNDARY_VECTORS`` or ``VECTOR_IN_INSTRUMENT_FOV``.

        Attributes :py:attr:`.direction_frame` is needed only if
        :py:attr:`.direction_vector_type` is ``REFERENCE_FRAME_AXIS`` or
        ``VECTOR_IN_REFERENCE_FRAME``.

        Attributes :py:attr:`.direction_vector_x` + :py:attr:`.direction_vector_y` +
        :py:attr:`.direction_vector_z` or :py:attr:`.direction_vector_ra` +
        :py:attr:`.direction_vector_dec` is needed only if
        :py:attr:`.direction_vector_type` is ``VECTOR_IN_INSTRUMENT_FOV`` or
        ``VECTOR_IN_REFERENCE_FRAME``.

    Raises
    ------
    CalculationRequiredAttr
        If :py:attr:`.target`, :py:attr:`.target_frame`
        and :py:attr:`.observer` are not provided.
    CalculationInvalidAttr
        If :py:attr:`.shape_1` is not ``ELLIPSOID`` or ``DSK``.

    '''

    def __init__(self, shape_1='ELLIPSOID', direction_vector_type='INSTRUMENT_BORESIGHT',
                 aberration_correction='CN', state_representation='RECTANGULAR',
                 **kwargs):

        self._required(['target', 'target_frame', 'observer'], kwargs)

        kwargs['calculation_type'] = 'SURFACE_INTERCEPT_POINT'
        kwargs['direction_vector_type'] = direction_vector_type
        kwargs['aberration_correction'] = aberration_correction
        kwargs['state_representation'] = state_representation

        if shape_1 in ['ELLIPSOID', 'DSK']:
            kwargs['shape_1'] = shape_1
        else:
            raise CalculationInvalidAttr('shape_1', shape_1, ['ELLIPSOID', 'DSK'])

        super().__init__(**kwargs)

class OsculatingElements(Calculation):
    '''Osculating elements calculation.

    Calculate the osculating elements of the orbit of a target body around a central body.

    The orbit may be elliptical, parabolic, or hyperbolic.

    Parameters
    ----------
    reference_frame: str, optional
        See: :py:attr:`reference_frame`

    Other Parameters
    ----------------
    kernels: str, int, [str or/and int]
        See: :py:attr:`.kernels`
    kernel_paths: str, [str]
        See: :py:attr:`.kernel_paths`
    times: str or [str]
        See: :py:attr:`.times`
    intervals: [str, str] or {'startTime': str, 'endTime': str} or [interval, ...]
        See: :py:attr:`.intervals`
    time_step: int
        See: :py:attr:`.time_step`
    time_step_units: str
        See: :py:attr:`.time_step_units`
    time_system: str
        See: :py:attr:`.time_system`
    time_format: str
        See: :py:attr:`.time_format`
    orbiting_body: str or int
        See: :py:attr:`.orbiting_body`
    center_body: str or int
        See: :py:attr:`.center_body`

    Raises
    ------
    CalculationRequiredAttr
        If :py:attr:`.orbiting_body` and :py:attr:`.center_body` are not provided.

    '''

    def __init__(self, reference_frame='J2000', **kwargs):

        self._required(['orbiting_body', 'center_body'], kwargs)

        kwargs['calculation_type'] = 'OSCULATING_ELEMENTS'
        kwargs['reference_frame'] = reference_frame

        super().__init__(**kwargs)

class TimeConversion(Calculation):
    '''Time conversion calculation.

    Convert times from one time system or format to another.

    Parameters
    ----------
    output_time_system: str, optional
        See: :py:attr:`output_time_system`
    output_time_format: str, optional
        See: :py:attr:`output_time_format`

    Other Parameters
    ----------------
    kernels: str, int, [str or/and int]
        See: :py:attr:`.kernels`
    kernel_paths: str, [str]
        See: :py:attr:`.kernel_paths`
    times: str or [str]
        See: :py:attr:`.times`
    intervals: [str, str] or {'startTime': str, 'endTime': str} or [interval, ...]
        See: :py:attr:`.intervals`
    time_step: int
        See: :py:attr:`.time_step`
    time_step_units: str
        See: :py:attr:`.time_step_units`
    time_system: str
        See: :py:attr:`.time_system`
    time_format: str
        See: :py:attr:`.time_format`
    output_sclk_id: str
        See: :py:attr:`.output_sclk_id`
    output_time_custom_format: str
        See: :py:attr:`.output_time_custom_format`

        Warnings
        --------
        Attributes :py:attr:`.output_sclk_id` is needed only if
        :py:attr:`output_time_system` is ``SPACECRAFT_CLOCK``.

        Attributes :py:attr:`.output_time_custom_format` is needed only if
        :py:attr:`output_time_format` is ``CUSTOM``.

    '''

    def __init__(self, output_time_system='UTC', output_time_format='CALENDAR', **kwargs):

        kwargs['calculation_type'] = 'TIME_CONVERSION'
        kwargs['output_time_system'] = output_time_system
        kwargs['output_time_format'] = output_time_format

        super().__init__(**kwargs)
